import warnings
from collections.abc import Mapping, Sequence
from fractions import Fraction
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import cvxpy.settings
import eth_abi.abi
import numpy as np
import web3
from cvxpy import Maximize, Parameter, Problem, Variable
from cvxpy.atoms.affine.binary_operators import multiply
from cvxpy.atoms.affine.bmat import bmat
from cvxpy.atoms.affine.hstack import hstack
from cvxpy.atoms.geo_mean import geo_mean
from cvxpy.error import SolverError
from eth_typing import ChecksumAddress

from degenbot import (
    AerodromeV2Pool,
    AerodromeV2PoolState,
    ArbitrageCalculationResult,
    Erc20Token,
    UniswapLpCycle,
    UniswapV2Pool,
    UniswapV2PoolState,
    get_checksum_address,
)
from degenbot.arbitrage.types import UniswapV2PoolSwapAmounts
from degenbot.exceptions.arbitrage import ArbitrageError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.logging import logger

DEBUG_VERIFY_CACHED_PROBLEM = False


class InvalidForwardAmount(ArbitrageError): ...


class Unprofitable(ArbitrageError): ...


class NoSolverSolution(ArbitrageError):
    def __init__(self, message: str = "Solver failed to converge on a solution.") -> None:
        self.message = message
        super().__init__(message=message)

    def __reduce__(self) -> tuple[Any, ...]:
        # Pickling will raise an exception if a reduction method is not defined
        return self.__class__, (self.message,)


def _build_convex_problem(num_pools: int) -> Problem:
    """
    Construct a DPP-compliant cvxpy problem with parameterized values for pool reserves. This
    allows the problem to be defined once at the class level, and rapidly re-solved at the instance
    level by updating the parameters for the specific pools and tokens being evaluated.

    The initial reserve, fee, and token decimal values are typical for the expected problem.

    ref: https://www.cvxpy.org/tutorial/dpp/index.html
    """

    class FakePool:
        token0: "FakeToken"
        token1: "FakeToken"
        tokens: tuple["FakeToken", "FakeToken"]
        fee_token0 = Fraction(3, 1000)
        fee_token1 = Fraction(3, 1000)

    class FakeToken: ...

    num_tokens = num_pools

    # Create unique fake pools & tokens to replicate the sorting and ordering mechanism used by
    # the instance method
    ordered_pools = [FakePool() for _ in range(num_pools)]
    ordered_tokens = [FakeToken() for _ in range(num_tokens)]

    for i, pool in enumerate(ordered_pools):
        if i == num_pools - 1:
            pool.token0 = ordered_tokens[-1]
            pool.token1 = ordered_tokens[0]
        else:
            pool.token0 = ordered_tokens[i]
            pool.token1 = ordered_tokens[i + 1]

        pool.tokens = pool.token0, pool.token1

    global_pool_index = {pool: i for i, pool in enumerate(ordered_pools)}
    global_token_index = {token: i for i, token in enumerate(ordered_tokens)}

    # The first pool holds token_1, token_2
    # The second holds token_2, token_3
    # ...
    # The last pool holds token_n, token_1 where token_n is equal to the number of tokens

    uncompressed_reserves = np.zeros(
        shape=(num_pools, num_tokens),
        dtype=np.float64,
    )
    for pool_index in range(num_pools):
        if pool_index == num_pools - 1:
            uncompressed_reserves[pool_index, pool_index] = 1 * 10**18
            uncompressed_reserves[pool_index, 0] = 1 * 10**18
        else:
            uncompressed_reserves[pool_index, pool_index] = 1 * 10**18
            uncompressed_reserves[pool_index, pool_index + 1] = 1 * 10**18

    # Identify the largest value to use as a common divisor for each token.
    token_compression_factors = [
        np.max(uncompressed_reserves[:, global_token_index[token]]) for token in ordered_tokens
    ]

    # SET UP PARAMETERS
    # Compress all pool reserves into a 0.0 - 1.0 value range by dividing by the compression
    # factor (via elementwise multiplication of the inverse)
    compressed_reserves_pre_swap = Parameter(
        shape=(num_pools, num_tokens),
        name="compressed_reserves_pre_swap",
        value=np.multiply(uncompressed_reserves, np.reciprocal(token_compression_factors)),
    )
    swap_fees = Parameter(
        shape=(num_pools, num_tokens),
        name="swap_fees",
        value=np.array(
            [
                [
                    (
                        pool.fee_token0
                        if token == pool.token0
                        else pool.fee_token1
                        if token in pool.tokens
                        else 0
                    )
                    for token in ordered_tokens
                ]
                for pool in ordered_pools
            ],
            dtype=np.float64,
        ),
    )
    pool_ks_pre_swap = Parameter(
        shape=len(ordered_pools),
        name="pool_ks_pre_swap",
        value=[
            geo_mean(
                hstack(
                    [
                        compressed_reserves_pre_swap[
                            global_pool_index[pool], global_token_index[token]
                        ]
                        for token in pool.tokens
                    ]
                )
            ).value
            for pool in ordered_pools
        ],
    )

    # SET UP VARIABLES
    initial_pool_deposit = Variable(name="initial_pool_deposit", nonneg=True)
    final_pool_withdrawal = Variable(name="final_pool_withdrawal", nonneg=True)
    # TODO: make profit token amounts forward token variables?
    forward_token_amount_variables = {
        token: Variable(name=f"forward_token_{global_token_index[token]}_amount", nonneg=True)
        for token in ordered_tokens[1:]  # skip the profit token
    }

    # SET UP PROBLEM
    _deposits: list[list[Literal[0] | Variable]]
    _deposits = [[0 for _ in ordered_tokens] for _ in ordered_pools]

    _withdrawals: list[list[Literal[0] | Variable]]
    _withdrawals = [[0 for _ in ordered_tokens] for _ in ordered_pools]

    for token, token_index in global_token_index.items():
        if token_index == 0:
            # Special case for profit token, which is deposited into first pool and
            # withdrawn from last pool
            deposit_pool_index = 0
            withdrawal_pool_index = -1
            deposit_variable = initial_pool_deposit
            withdrawal_variable = final_pool_withdrawal
        else:
            # Each forward token is withdrawn from the pool at the previous position
            # e.g. token 1 withdrawn from pool 0, and deposited at the next pool
            deposit_pool_index = token_index
            withdrawal_pool_index = token_index - 1
            deposit_variable = withdrawal_variable = forward_token_amount_variables[token]

        _withdrawals[withdrawal_pool_index][token_index] = withdrawal_variable
        _deposits[deposit_pool_index][token_index] = deposit_variable

    deposits = bmat(_deposits)
    withdrawals = bmat(_withdrawals)

    compressed_reserves_post_swap = (
        compressed_reserves_pre_swap + deposits - withdrawals - multiply(swap_fees, deposits)
    )

    pool_ks_post_swap = [
        geo_mean(
            hstack(
                [
                    compressed_reserves_post_swap[
                        global_pool_index[pool], global_token_index[token]
                    ]
                    for token in pool.tokens
                ]
            )
        )
        for pool in ordered_pools
    ]

    constraints = []

    # Pool invariants (x*y=k)
    constraints.extend(
        [
            pool_ks_pre_swap[global_pool_index[pool]] <= pool_ks_post_swap[global_pool_index[pool]]
            for pool in ordered_pools
        ]
    )

    problem = Problem(
        objective=Maximize(final_pool_withdrawal - initial_pool_deposit),
        constraints=constraints,
    )
    assert problem.is_dcp(dpp=True)  # type: ignore[call-arg]
    problem.solve(solver="CLARABEL")
    return problem


type Pool = UniswapV2Pool | AerodromeV2Pool
type PoolState = UniswapV2PoolState | AerodromeV2PoolState
type SwapAmount = UniswapV2PoolSwapAmounts


class _UniswapMultiPoolCycleTesting(UniswapLpCycle):
    swap_pools: tuple[Pool, ...]
    convex_problems: ClassVar[
        dict[
            int,  # number of pools
            Problem,
        ]
    ] = {
        3: _build_convex_problem(num_pools=3),
        4: _build_convex_problem(num_pools=4),
    }

    def _calculate(  # type: ignore[override]
        self,
        state_overrides: Mapping[Pool, PoolState] | None = None,
    ) -> ArbitrageCalculationResult:
        """
        Calculate the optimal arbitrage profit using the maximum input as an upper bound.
        """

        pool_states: Mapping[Pool, PoolState]
        pool_states = {pool: pool.state for pool in self.swap_pools}

        if state_overrides is not None:
            pool_states.update(state_overrides)

        def v2_only_calc(
            pools: Sequence[Pool],
            pool_states: Mapping[Pool, PoolState],
        ) -> ArbitrageCalculationResult:
            """
            Calculate the optimal arbitrage for a sequence of Uniswap V2 (or compatible) pools of
            arbitrary length.
            """

            def get_token_balance_at_pool(
                token: Erc20Token,
                pool: Pool,
                state_override: PoolState | None = None,
            ) -> Fraction:
                if token not in pool.tokens:
                    return Fraction(0)

                _state = state_override or pool.state
                return Fraction(
                    _state.reserves_token0 if token == pool.token0 else _state.reserves_token1,
                    10**token.decimals,
                )

            def order_tokens(pools: Sequence[Pool]) -> tuple[Erc20Token, ...]:
                ordered_tokens: list[Erc20Token] = [self.input_token]

                for pool in pools:
                    for token in pool.tokens:
                        if token not in ordered_tokens:
                            ordered_tokens.append(token)

                assert len(ordered_tokens) == len(pools), f"{ordered_tokens=}, {pools=}"
                return tuple(ordered_tokens)

            # Reuse the pre-compiled problem
            problem = type(self).convex_problems[len(self.swap_pools)]

            # TODO: review if pools should be dynamically sorted, or filtered pre-check and assumed
            # to be passed in order
            ordered_pools = pools
            ordered_tokens = order_tokens(pools)

            global_pool_index = {pool: i for i, pool in enumerate(ordered_pools)}
            global_token_index = {token: i for i, token in enumerate(ordered_tokens)}

            uncompressed_reserves = np.array(
                [
                    [
                        get_token_balance_at_pool(
                            token=token,
                            pool=pool,
                            state_override=pool_states.get(pool),
                        )
                        for token in ordered_tokens
                    ]
                    for pool in ordered_pools
                ],
                dtype=np.float64,
            )

            # Identify the largest value to use as a common divisor for each token.
            token_compression_factors = [
                np.max(uncompressed_reserves[:, global_token_index[token]])
                for token in ordered_tokens
            ]

            # SET UP PARAMETERS
            assert len(problem.param_dict) == 3  # noqa: PLR2004
            swap_fees = problem.param_dict["swap_fees"]
            compressed_reserves_pre_swap = problem.param_dict["compressed_reserves_pre_swap"]
            pool_ks_pre_swap = problem.param_dict["pool_ks_pre_swap"]

            if TYPE_CHECKING:
                assert isinstance(swap_fees, cvxpy.Parameter)
                assert isinstance(compressed_reserves_pre_swap, cvxpy.Parameter)
                assert isinstance(pool_ks_pre_swap, cvxpy.Parameter)

            swap_fees.save_value(
                np.array(
                    [
                        [
                            pool.fee_token0
                            if token == pool.token0
                            else pool.fee_token1
                            if token in pool.tokens
                            else 0
                            for token in ordered_tokens
                        ]
                        for pool in ordered_pools
                    ],
                    dtype=np.float64,
                ),
            )
            compressed_reserves_pre_swap.save_value(
                np.multiply(uncompressed_reserves, np.reciprocal(token_compression_factors))
            )
            pool_ks_pre_swap.save_value(
                [
                    geo_mean(
                        hstack(
                            [
                                compressed_reserves_pre_swap[
                                    global_pool_index[pool], global_token_index[token]
                                ]
                                for token in pool.tokens
                            ]
                        )
                    ).value
                    for pool in ordered_pools
                ]
            )

            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    problem.solve(solver=cvxpy.CLARABEL)
            except SolverError as exc:
                raise ArbitrageError(message=f"Solver error: {exc}") from None

            if problem.status not in cvxpy.settings.SOLUTION_PRESENT:
                raise NoSolverSolution

            if problem.value <= 0:
                raise Unprofitable

            if DEBUG_VERIFY_CACHED_PROBLEM:
                try:
                    new_problem = Problem(
                        objective=problem.objective,
                        constraints=problem.constraints,
                    )
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        new_problem.solve(solver=cvxpy.CLARABEL)
                except SolverError:
                    raise ArbitrageError(message="Solver error") from None
                else:
                    if new_problem.value != problem.value:
                        result_percent_difference = (
                            100
                            * abs(new_problem.value - problem.value)
                            / ((new_problem.value + problem.value) / 2)
                        )
                        logger.error(
                            f"Cached problem result ({problem.value}) within {result_percent_difference:.2f}% of fresh result ({new_problem.value})"  # noqa: E501
                        )
                        err_msg = "Result mismatch"
                        raise ValueError(err_msg)

            amounts: list[UniswapV2PoolSwapAmounts] = []
            initial_pool_deposit = problem.var_dict["initial_pool_deposit"]
            initial_amount_in = int(
                initial_pool_deposit.value
                * token_compression_factors[global_token_index[self.input_token]]
                * 10**self.input_token.decimals
            )
            for i, pool in enumerate(ordered_pools):
                if i == 0:
                    amount_in = initial_amount_in
                    token_in = self.input_token

                if amount_in == 0:
                    raise ArbitrageError(message="Zero amount swap")

                zero_for_one = token_in == pool.token0
                token_out = pool.token1 if zero_for_one else pool.token0

                pool_state = pool_states[pool]

                match pool, pool_state:
                    case AerodromeV2Pool(), AerodromeV2PoolState():
                        try:
                            amount_out = pool.calculate_tokens_out_from_tokens_in(
                                token_in=token_in,
                                token_in_quantity=amount_in,
                                override_state=pool_state,
                            )
                        except (EVMRevertError, LiquidityPoolError) as e:
                            raise ArbitrageError from e

                    case UniswapV2Pool(), UniswapV2PoolState():
                        try:
                            amount_out = pool.calculate_tokens_out_from_tokens_in(
                                token_in=token_in,
                                token_in_quantity=amount_in,
                                override_state=pool_state,
                            )
                        except (EVMRevertError, LiquidityPoolError) as e:
                            raise ArbitrageError from e

                if amount_out == 0:
                    raise ArbitrageError(message="Zero amount swap")

                amounts.append(
                    UniswapV2PoolSwapAmounts(
                        pool=pool.address,
                        amounts_in=(amount_in, 0) if zero_for_one else (0, amount_in),
                        amounts_out=(0, amount_out) if zero_for_one else (amount_out, 0),
                    )
                )

                amount_in = amount_out
                token_in = token_out

            if (best_profit := amount_out - initial_amount_in) <= 0:
                raise Unprofitable

            newest_state_block = None
            if not state_overrides:
                pool_state_blocks = tuple(
                    block for pool in self.swap_pools if (block := pool.state.block) is not None
                )
                if len(pool_state_blocks) == len(self.swap_pools):
                    newest_state_block = max(pool_state_blocks)

            return ArbitrageCalculationResult(
                id=self.id,
                input_token=self.input_token,
                profit_token=self.input_token,
                input_amount=initial_amount_in,
                profit_amount=best_profit,
                swap_amounts=tuple(amounts),
                state_block=newest_state_block,
            )

        if all(
            pool for pool in self.swap_pools if isinstance(pool, AerodromeV2Pool | UniswapV2Pool)
        ):
            return v2_only_calc(
                pools=self.swap_pools,
                pool_states=pool_states,
            )

        err = "One or more pools is not supported by this arbitrage strategy."
        raise ValueError(err)

    def generate_payloads(  # type: ignore[override]
        self,
        from_address: ChecksumAddress | str,
        pool_swap_amounts: Sequence[UniswapV2PoolSwapAmounts],
    ) -> list[tuple[ChecksumAddress, bytes, bool]]:
        from_address = get_checksum_address(from_address)

        """
        PAYLOAD DEFINITION FROM CONTRACT

        struct Payload:
            target: address
            calldata: Bytes[MAX_PAYLOAD_BYTES]
            will_callback: bool
        """

        def _generate_v2_v2_v2_payloads() -> list[tuple[ChecksumAddress, bytes, bool]]:
            return [
                (
                    # Swap at final pool, send profit token back to contract
                    pool_swap_amounts[2].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[2].amounts_out,
                            from_address,
                            b"x",  # <--- trigger callback
                        ),
                    ),
                    True,
                ),
                (
                    # Transfer token to pay first pool
                    self.input_token.address,
                    web3.Web3.keccak(text="transfer(address,uint256)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "address",
                            "uint256",
                        ),
                        args=(
                            pool_swap_amounts[0].pool,
                            max(pool_swap_amounts[0].amounts_in),
                        ),
                    ),
                    False,
                ),
                (
                    # First pool swap
                    pool_swap_amounts[0].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[0].amounts_out,
                            pool_swap_amounts[1].pool,
                            b"",
                        ),
                    ),
                    False,
                ),
                (
                    # Final swap
                    pool_swap_amounts[1].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[1].amounts_out,
                            pool_swap_amounts[2].pool,
                            b"",
                        ),
                    ),
                    False,
                ),
            ]

        def _generate_v2_v2_v2_v2_payloads() -> list[tuple[ChecksumAddress, bytes, bool]]:
            return [
                (
                    # Swap at final pool, send profit token back to contract
                    pool_swap_amounts[3].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[3].amounts_out,
                            from_address,
                            b"x",  # <--- trigger callback
                        ),
                    ),
                    True,
                ),
                (
                    # Transfer token to pay first pool
                    self.input_token.address,
                    web3.Web3.keccak(text="transfer(address,uint256)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "address",
                            "uint256",
                        ),
                        args=(
                            pool_swap_amounts[0].pool,
                            max(pool_swap_amounts[0].amounts_in),
                        ),
                    ),
                    False,
                ),
                (
                    # First pool swap
                    pool_swap_amounts[0].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[0].amounts_out,
                            pool_swap_amounts[1].pool,
                            b"",
                        ),
                    ),
                    False,
                ),
                (
                    # Second pool swap
                    pool_swap_amounts[1].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[1].amounts_out,
                            pool_swap_amounts[2].pool,
                            b"",
                        ),
                    ),
                    False,
                ),
                (
                    # Third pool swap
                    pool_swap_amounts[2].pool,
                    web3.Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                    + eth_abi.abi.encode(
                        types=(
                            "uint256",
                            "uint256",
                            "address",
                            "bytes",
                        ),
                        args=(
                            *pool_swap_amounts[2].amounts_out,
                            pool_swap_amounts[3].pool,
                            b"",
                        ),
                    ),
                    False,
                ),
            ]

        match self.swap_pools:
            case (
                AerodromeV2Pool() | UniswapV2Pool(),
                AerodromeV2Pool() | UniswapV2Pool(),
                AerodromeV2Pool() | UniswapV2Pool(),
            ):
                return _generate_v2_v2_v2_payloads()
            case (
                AerodromeV2Pool() | UniswapV2Pool(),
                AerodromeV2Pool() | UniswapV2Pool(),
                AerodromeV2Pool() | UniswapV2Pool(),
                AerodromeV2Pool() | UniswapV2Pool(),
            ):
                return _generate_v2_v2_v2_v2_payloads()
            case _:
                err = f"Could not identify pool types {self.swap_pools}"
                raise ValueError(err)



from . import abi
from .managers import PancakeV2PoolManager, PancakeV3PoolManager
from .pools import PancakeV2Pool, PancakeV3Pool

__all__ = (
    "PancakeV2Pool",
    "PancakeV2PoolManager",
    "PancakeV3Pool",
    "PancakeV3PoolManager",
    "abi",
)



from degenbot import get_checksum_address

CURVE_V1_REGISTRY_ADDRESS = get_checksum_address("0x90E00ACe148ca3b23Ac1bC8C240C2a7Dd9c2d7f5")
CURVE_V1_FACTORY_ADDRESS = get_checksum_address("0x127db66E7F0b16470Bec194d0f496F9Fa065d0A9")
CURVE_V1_METAREGISTRY_ADDRESS = get_checksum_address("0xF98B45FA17DE75FB1aD0e7aFD971b0ca00e379fC")

BROKEN_CURVE_V1_POOLS = frozenset(
    get_checksum_address(pool_address)
    for pool_address in (
        "0x110cc323ca53d622469EdD217387E2E6B33F1dF5",
        "0x1F71f05CF491595652378Fe94B7820344A551B8E",
        "0x28B0Cf1baFB707F2c6826d10caf6DD901a6540C5",
        "0x3685646651FCcC80e7CCE7Ee24c5f47Ed9b434ac",
        "0x84997FAFC913f1613F51Bb0E2b5854222900514B",
        "0x9CA41a2DaB3CEE15308998868ca644e2e3be5C59",
        "0xA77d09743F77052950C4eb4e6547E9665299BecD",
        "0xD652c40fBb3f06d6B58Cb9aa9CFF063eE63d465D",
    )
)



from degenbot.sushiswap.pools import SushiswapV2Pool, SushiswapV3Pool
from degenbot.uniswap.managers import UniswapV2PoolManager, UniswapV3PoolManager


class SushiswapV2PoolManager(UniswapV2PoolManager, pool_factory=SushiswapV2Pool):
    type Pool = SushiswapV2Pool


class SushiswapV3PoolManager(UniswapV3PoolManager, pool_factory=SushiswapV3Pool):
    type Pool = SushiswapV3Pool



import contextlib
from typing import TYPE_CHECKING, Self

import degenbot.exceptions
from degenbot import get_checksum_address
from degenbot.exceptions.registry import RegistryAlreadyInitialized
from degenbot.types.aliases import ChainId

if TYPE_CHECKING:
    from eth_typing import ChecksumAddress

    from degenbot.erc20 import Erc20Token


class TokenRegistry:
    instance: Self | None = None

    @classmethod
    def get_instance(cls) -> Self | None:
        return cls.instance

    def __init__(self) -> None:
        if self.__class__.instance is not None:
            raise RegistryAlreadyInitialized(
                message="A registry has already been initialized. Access it using the get_instance() class method"  # noqa:E501
            )
        self.__class__.instance = self

        self._all_tokens: dict[
            tuple[
                int,  # chain ID
                ChecksumAddress,  # token address
            ],
            Erc20Token,
        ] = {}

    def get(self, token_address: str, chain_id: ChainId) -> "Erc20Token | None":
        return self._all_tokens.get(
            (chain_id, get_checksum_address(token_address)),
        )

    def add(self, token_address: str, chain_id: ChainId, token: "Erc20Token") -> None:
        token_address = get_checksum_address(token_address)
        if self.get(token_address=token_address, chain_id=chain_id):
            raise degenbot.exceptions.DegenbotValueError(message="Token is already registered")
        self._all_tokens[(chain_id, token_address)] = token

    def remove(self, token_address: str, chain_id: ChainId) -> None:
        token_address = get_checksum_address(token_address)

        with contextlib.suppress(KeyError):
            del self._all_tokens[(chain_id, token_address)]



from fractions import Fraction

from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool


class PancakeV2Pool(UniswapV2Pool):
    FEE = Fraction(25, 10000)
    RESERVES_STRUCT_TYPES = (
        "uint112",
        "uint112",
        "uint32",
    )  # type:ignore[assignment]


class PancakeV3Pool(UniswapV3Pool):
    SLOT0_STRUCT_TYPES = (
        "uint160",
        "int24",
        "uint16",
        "uint16",
        "uint16",
        "uint32",
        "bool",
    )



from collections.abc import Callable
from fractions import Fraction
from typing import Literal

from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.functions import raise_if_invalid_uint256


def general_calc_d(
    x0: int,
    y: int,
) -> int:
    return (3 * x0 * ((y * y) // 10**18)) // 10**18 + ((((x0 * x0) // 10**18) * x0) // 10**18)


def general_calc_k(
    balance_0: int,
    balance_1: int,
    decimals_0: int,
    decimals_1: int,
) -> int:
    _x = balance_0 * 10**18 // decimals_0
    _y = balance_1 * 10**18 // decimals_1
    _a = (_x * _y) // 10**18
    _b = (_x * _x) // 10**18 + (_y * _y) // 10**18
    raise_if_invalid_uint256(_a * _b)
    return _a * _b // 10**18  # x^3*y + y^3*x >= k


def general_calc_exact_in_stable(
    amount_in: int,
    token_in: Literal[0, 1],
    reserves0: int,
    reserves1: int,
    decimals0: int,
    decimals1: int,
    fee: Fraction,
    k_func: Callable[
        [int, int, int, int],
        int,
    ],
    get_y_func: Callable[
        [int, int, int, int, int],
        int,
    ],
) -> int:
    """
    Calculate the amount out for an exact input to a Solidly stable pool (invariant
    y*x^3 + x*y^3 >= k).

    This function is generic and requires a callable k() and get_y()
    """

    if token_in not in (0, 1):  # pragma: no cover
        raise DegenbotValueError(message="Invalid token_in identifier")

    try:
        amount_in_after_fee = amount_in - amount_in * fee.numerator // fee.denominator

        xy = k_func(reserves0, reserves1, decimals0, decimals1)

        scaled_reserves_0 = (reserves0 * 10**18) // decimals0
        scaled_reserves_1 = (reserves1 * 10**18) // decimals1

        if token_in == 0:
            reserves_a, reserves_b = scaled_reserves_0, scaled_reserves_1
            amount_in_after_fee = (amount_in_after_fee * 10**18) // decimals0
        else:
            reserves_a, reserves_b = scaled_reserves_1, scaled_reserves_0
            amount_in_after_fee = (amount_in_after_fee * 10**18) // decimals1

        y = reserves_b - get_y_func(
            amount_in_after_fee + reserves_a, xy, reserves_b, decimals0, decimals1
        )
        return (y * (decimals1 if token_in == 0 else decimals0)) // 10**18
    except ZeroDivisionError:
        # Pools with very low reserves can throw division by zero errors because _d() returns 0
        raise EVMRevertError(error="Division by zero") from None


def general_calc_exact_in_volatile(
    amount_in: int,
    token_in: Literal[0, 1],
    reserves0: int,
    reserves1: int,
    fee: Fraction,
) -> int:
    """
    Calculate the amount out for an exact input to a Solidly volatile pool (invariant x*y>=k).
    """

    if token_in not in (0, 1):  # pragma: no cover
        raise DegenbotValueError(message="Invalid token_in identifier")

    amount_in_after_fee = amount_in - amount_in * fee.numerator // fee.denominator
    reserves_a, reserves_b = (reserves0, reserves1) if token_in == 0 else (reserves1, reserves0)
    return (amount_in_after_fee * reserves_b) // (reserves_a + amount_in_after_fee)



import json

import pytest
from click.testing import CliRunner

from degenbot import __version__
from degenbot.cli import cli


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


def test_cli_help(runner: CliRunner):
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0


def test_cli_version(runner: CliRunner):
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.output


def test_cli_config_show_default(runner: CliRunner):
    result = runner.invoke(cli, ["config", "show"])
    assert result.exit_code == 0
    assert "[database]" in result.output


def test_cli_config_show_json(runner: CliRunner):
    result = runner.invoke(cli, ["config", "show", "--json"])
    assert result.exit_code == 0
    assert json.loads(result.output)


def test_cli_config_show_toml(runner: CliRunner):
    result = runner.invoke(cli, ["config", "show", "--toml"])
    assert result.exit_code == 0
    assert "[database]" in result.output


# TODO: create fake database for reset/upgrade tests


def test_cli_database_reset(runner: CliRunner):
    result = runner.invoke(cli, ["database", "reset"], input="n")
    assert result.exit_code == 1

    result = runner.invoke(cli, ["database", "reset"], input="")
    assert result.exit_code == 1


def test_cli_database_upgrade(runner: CliRunner):
    result = runner.invoke(cli, ["database", "upgrade"], input="n")
    assert result.exit_code == 1

    result = runner.invoke(cli, ["database", "upgrade"], input="")
    assert result.exit_code == 1



import functools
from typing import Annotated

from pydantic import Field, validate_call

from degenbot.constants import MAX_INT16, MAX_UINT256
from degenbot.functions import evm_divide
from degenbot.uniswap.v4_libraries import bit_math
from degenbot.uniswap.v4_libraries._config import V4_LIB_CACHE_SIZE
from degenbot.validation.evm_values import ValidatedInt24, ValidatedUint160

MIN_TICK = -887272
MAX_TICK = 887272
MIN_TICK_SPACING = 1
MAX_TICK_SPACING = MAX_INT16
MIN_SQRT_PRICE = 4295128739
MAX_SQRT_PRICE = 1461446703485210103287273052203988822378723970342
MAX_SQRT_PRICE_MINUS_MIN_SQRT_PRICE_MINUS_ONE = (
    1461446703485210103287273052203988822378723970342 - 4295128739 - 1
)


# Magic number represents the minimum value of the error when approximating log_sqrt10001(x),
# when sqrtPrice is from the range (2^-64, 2^64). This is safe as MIN_SQRT_PRICE is more than
# 2^-64. If MIN_SQRT_PRICE is changed, this may need to be changed too
MIN_ERROR = 291339464771989622907027621153398088495

# Magic number represents the ceiling of the maximum value of the error when
# approximating log_sqrt10001(x)
MAX_ERROR = 3402992956809132418596140100660247210


type ValidatedTick = Annotated[int, Field(strict=True, ge=MIN_TICK, le=MAX_TICK)]
type ValidatedSqrtPrice = Annotated[int, Field(strict=True, ge=MIN_SQRT_PRICE, le=MAX_SQRT_PRICE)]


@validate_call(validate_return=True)
def max_usable_tick(tick_spacing: ValidatedInt24) -> ValidatedInt24:
    """
    Given a tickSpacing, compute the maximum usable tick
    """

    return (MAX_TICK // tick_spacing) * tick_spacing


@validate_call(validate_return=True)
def min_usable_tick(tick_spacing: ValidatedInt24) -> ValidatedInt24:
    """
    Given a tickSpacing, compute the minimum usable tick
    """

    return evm_divide(MIN_TICK, tick_spacing) * tick_spacing


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_sqrt_price_at_tick(tick: ValidatedTick) -> ValidatedUint160:
    """
    Calculates sqrt(1.0001^tick) * 2^96, a fixed point Q64.96 number representing the sqrt of the
    price of the two assets (currency1/currency0) at the given tick.

    Raises exception if |tick| > max tick.
    """

    # Use abs instead of reimplementing the Solidity contract's inline Yul
    abs_tick = abs(tick)

    # The tick is decomposed into bits, and for each bit with index i that is set, the product of
    # 1/sqrt(1.0001^(2^i)) is calculated (using Q128.128). The constants used for this calculation
    # are rounded to the nearest integer

    # Equivalent to:
    #     price = absTick & 0x1 != 0 ?
    #         0xfffcb933bd6fad37aa2d162d1a594001 :
    #         0x100000000000000000000000000000000;
    #     or price = int(2**128 / sqrt(1.0001)) if (absTick & 0x1) else 1 << 128

    price = (1 << 128) ^ (((1 << 128) ^ 340265354078544963557816517032075149313) * (abs_tick & 1))

    for tick_mask, ratio_multiplier in (
        (2, 340248342086729790484326174814286782778),
        (4, 340214320654664324051920982716015181260),
        (8, 340146287995602323631171512101879684304),
        (16, 340010263488231146823593991679159461444),
        (32, 339738377640345403697157401104375502016),
        (64, 339195258003219555707034227454543997025),
        (128, 338111622100601834656805679988414885971),
        (256, 335954724994790223023589805789778977700),
        (512, 331682121138379247127172139078559817300),
        (1024, 323299236684853023288211250268160618739),
        (2048, 307163716377032989948697243942600083929),
        (4096, 277268403626896220162999269216087595045),
        (8192, 225923453940442621947126027127485391333),
        (16384, 149997214084966997727330242082538205943),
        (32768, 66119101136024775622716233608466517926),
        (65536, 12847376061809297530290974190478138313),
        (131072, 485053260817066172746253684029974020),
        (262144, 691415978906521570653435304214168),
        (524288, 1404880482679654955896180642),
    ):
        if abs_tick & tick_mask != 0:
            price = (price * ratio_multiplier) >> 128

    if tick > 0:
        price = MAX_UINT256 // price

    # This divides by 1<<32 rounding up to go from a Q128.128 to a Q128.96.
    # Then downcast because the result always fits within 160 bits due to our tick input constraint.
    # Round up in the division so getTickAtSqrtPrice of the output price is always consistent
    # `sub(shl(32, 1), 1)` is `type(uint32).max`
    # `price + type(uint32).max` will not overflow because `price` fits in 192 bits
    return (price + ((1 << 32) - 1)) >> 32


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_tick_at_sqrt_price(sqrt_price_x96: ValidatedSqrtPrice) -> ValidatedTick:
    """
    Calculates the greatest tick value such that getSqrtPriceAtTick(tick) <= sqrtPriceX96

    @dev raises exception if sqrt_price_x96 is below MIN_SQRT_PRICE or above MAX_SQRT_PRICE.
    """

    price = sqrt_price_x96 << 32
    msb = bit_math.most_significant_bit(price)
    r = price >> msb - 127 if msb >= 128 else price << 127 - msb  # noqa: PLR2004
    log_2 = (msb - 128) << 64

    for factor in (63, 62, 61, 60, 59, 58, 57, 56, 55, 54, 53, 52, 51):
        r = (r * r) >> 127
        f = r >> 128
        log_2 |= f << factor
        r >>= f

    r = (r * r) >> 127
    f = r >> 128
    log_2 |= f << 50

    log_sqrt10001 = log_2 * 255738958999603826347141  # Q22.128 number

    tick_low = (log_sqrt10001 - MAX_ERROR) >> 128
    tick_high = (log_sqrt10001 + MIN_ERROR) >> 128

    return (
        tick_low
        if tick_low == tick_high
        else (tick_high if get_sqrt_price_at_tick(tick_high) <= sqrt_price_x96 else tick_low)
    )



from degenbot.exceptions import DegenbotError


class Erc20TokenError(DegenbotError):
    """
    Exception raised inside ERC-20 token helpers.
    """


class NoPriceOracle(Erc20TokenError):
    """
    Raised when `.price` is called on a token without a price oracle.
    """

    def __init__(self) -> None:
        super().__init__(message="Token does not have a price oracle.")



from typing import Any

from eth_typing import ChecksumAddress

from degenbot.logging import logger
from degenbot.pancakeswap.pools import PancakeV2Pool, PancakeV3Pool
from degenbot.uniswap.managers import UniswapV2PoolManager, UniswapV3PoolManager


class PancakeV2PoolManager(UniswapV2PoolManager, pool_factory=PancakeV2Pool): ...


class PancakeV3PoolManager(UniswapV3PoolManager, pool_factory=PancakeV3Pool):
    def _build_pool(
        self,
        pool_address: ChecksumAddress,
        *,
        silent: bool,
        pool_class_kwargs: dict[str, Any] | None,
    ) -> PancakeV3Pool:
        if pool_class_kwargs is None:
            pool_class_kwargs = {}

        if self._snapshot is not None:
            pool = PancakeV3Pool(
                address=pool_address,
                tick_bitmap=self._snapshot.tick_bitmap(pool_address),
                tick_data=self._snapshot.tick_data(pool_address),
                silent=silent,
                **pool_class_kwargs,
            )
        else:
            logger.info("Initializing pool without liquidity snapshot")
            pool = PancakeV3Pool(
                address=pool_address,
                silent=silent,
                **pool_class_kwargs,
            )
        return pool



# ruff: noqa: E501

from fractions import Fraction
from threading import Lock
from typing import cast
from weakref import WeakSet

import cvxpy
import cvxpy.settings
import numpy as np
import pytest
from cvxpy.atoms.affine.binary_operators import multiply as cvxpy_multiply
from cvxpy.atoms.affine.bmat import bmat as cvxpy_bmat
from cvxpy.atoms.affine.sum import sum as cvxpy_sum
from cvxpy.atoms.geo_mean import geo_mean

from degenbot import (
    AnvilFork,
    Erc20Token,
    UniswapV2Pool,
    UniswapV2PoolState,
    get_checksum_address,
    set_web3,
)
from degenbot.arbitrage.uniswap_multipool_cycle_testing import _UniswapMultiPoolCycleTesting
from degenbot.constants import ZERO_ADDRESS

WBTC_ADDRESS = "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
WETH_ADDRESS = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
LINK_ADDRESS = "0x514910771AF9Ca656af840dff83E8264EcF986CA"
WBTC_WETH_V2_POOL_ADDRESS = "0xBb2b8038a1640196FbE3e38816F3e67Cba72D940"
WBTC_WETH_V3_POOL_ADDRESS = "0xCBCdF9626bC03E24f779434178A73a0B4bad62eD"


class MockLiquidityPool(UniswapV2Pool):
    def __init__(self) -> None:
        self._state = UniswapV2PoolState(
            address=ZERO_ADDRESS,
            reserves_token0=0,
            reserves_token1=0,
            block=None,
        )
        self._state_lock = Lock()
        self._subscribers = WeakSet()


@pytest.fixture
def wbtc_token(fork_mainnet_full: AnvilFork) -> Erc20Token:
    set_web3(fork_mainnet_full.w3)
    return Erc20Token(WBTC_ADDRESS)


@pytest.fixture
def weth_token(fork_mainnet_full: AnvilFork) -> Erc20Token:
    set_web3(fork_mainnet_full.w3)
    return Erc20Token(WETH_ADDRESS)


@pytest.fixture
def link_token(fork_mainnet_full: AnvilFork) -> Erc20Token:
    set_web3(fork_mainnet_full.w3)
    return Erc20Token(LINK_ADDRESS)


@pytest.fixture
def weth_base_token(fork_base_full: AnvilFork) -> Erc20Token:
    set_web3(fork_base_full.w3)
    return Erc20Token("0x4200000000000000000000000000000000000006")


@pytest.fixture
def xxx_token(fork_base_full: AnvilFork) -> Erc20Token:
    set_web3(fork_base_full.w3)
    return Erc20Token("0x09C07E80bFeEd81130498516F5C07aA0715794Bb")


@pytest.fixture
def wbtc_pool_a(wbtc_token, weth_token) -> MockLiquidityPool:
    lp = MockLiquidityPool()
    lp.name = "WBTC-WETH (V2, 0.30%)"
    lp.address = get_checksum_address("0xBb2b8038a1640196FbE3e38816F3e67Cba72D940")
    lp.factory = get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")
    lp.fee_token0 = Fraction(3, 1000)
    lp.fee_token1 = Fraction(3, 1000)
    lp._state = UniswapV2PoolState(
        address=lp.address,
        block=None,
        reserves_token0=9000000000,
        reserves_token1=2100000000000000000000,
    )
    lp.token0 = wbtc_token
    lp.token1 = weth_token
    return lp


@pytest.fixture
def wbtc_pool_b(wbtc_token, weth_token) -> MockLiquidityPool:
    lp = MockLiquidityPool()
    lp.name = "WBTC-WETH (V2, 0.30%)"
    lp.address = get_checksum_address("0xBb2b8038a1640196FbE3e38816F3e67Cba72D941")
    lp.factory = get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")
    lp.fee_token0 = Fraction(3, 1000)
    lp.fee_token1 = Fraction(3, 1000)
    lp._state = UniswapV2PoolState(
        address=lp.address,
        block=None,
        reserves_token0=9250000000,
        reserves_token1=2100000000000000000000,
    )
    lp.token0 = wbtc_token
    lp.token1 = weth_token
    return lp


@pytest.fixture
def test_pool_a(xxx_token, weth_base_token) -> MockLiquidityPool:
    lp = MockLiquidityPool()
    lp.name = ""
    lp.address = get_checksum_address("0x214356Cc4aAb907244A791CA9735292860490D5A")
    lp.factory = get_checksum_address("0x420DD381b31aEf6683db6B902084cB0FFECe40Da")
    lp.fee_token0 = Fraction(3, 1000)
    lp.fee_token1 = Fraction(3, 1000)
    lp._state = UniswapV2PoolState(
        address=lp.address,
        block=None,
        reserves_token0=19643270033194347,
        reserves_token1=406789256841523130269,
    )
    lp.token0 = weth_base_token
    lp.token1 = xxx_token
    return lp


@pytest.fixture
def test_pool_b(xxx_token, weth_base_token) -> MockLiquidityPool:
    lp = MockLiquidityPool()
    lp.name = ""
    lp.address = get_checksum_address("0x404E927b203375779a6aBD52A2049cE0ADf6609B")
    lp.factory = get_checksum_address("0x8909Dc15e40173Ff4699343b6eB8132c65e18eC6")
    lp.fee_token0 = Fraction(3, 1000)
    lp.fee_token1 = Fraction(3, 1000)
    lp._state = UniswapV2PoolState(
        address=lp.address,
        block=None,
        reserves_token0=880450452482804609420,
        reserves_token1=18733831498401825763565574,
    )
    lp.token0 = weth_base_token
    lp.token1 = xxx_token
    return lp


def test_2pool_uniswap_v2_decimal_corrected(
    wbtc_pool_a: MockLiquidityPool,
    wbtc_pool_b: MockLiquidityPool,
    weth_token: Erc20Token,
):
    profit_token = weth_token
    pool_a_roe = wbtc_pool_a.get_absolute_exchange_rate(token=profit_token)
    pool_b_roe = wbtc_pool_b.get_absolute_exchange_rate(token=profit_token)

    if pool_a_roe > pool_b_roe:
        pool_hi = wbtc_pool_a
        pool_lo = wbtc_pool_b
    else:
        pool_hi = wbtc_pool_b
        pool_lo = wbtc_pool_a

    num_pools = 2
    num_tokens = 2

    # Indices are arbitrary but must be consistent so amounts are consistent between pools in the
    # various arrays
    pool_hi_index, pool_lo_index = 0, 1

    token0_decimals = pool_hi.token0.decimals
    token1_decimals = pool_hi.token1.decimals

    forward_token = pool_hi.token1 if pool_hi.token0 == profit_token else pool_hi.token0
    forward_token_index = 1 if pool_hi.token0 == profit_token else 0
    profit_token_index = 0 if pool_hi.token0 == profit_token else 1
    assert forward_token_index != profit_token_index

    pool_hi_fees = [pool_hi.fee_token0, pool_hi.fee_token1]
    pool_lo_fees = [pool_lo.fee_token0, pool_lo.fee_token1]
    fee_multiplier = cvxpy_bmat(
        (
            pool_hi_fees,
            pool_lo_fees,
        )
    )

    # Identify the largest value to use as a common divisor.
    compression_factor = max(
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals),
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals),
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals),
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals),
    )

    # Compress all pool reserves into a 0.0 - 1.0 value range
    _compressed_starting_reserves_pool_hi = (
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals) / compression_factor,
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals) / compression_factor,
    )
    _compressed_starting_reserves_pool_lo = (
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals) / compression_factor,
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals) / compression_factor,
    )
    compressed_reserves_pre_swap = cvxpy.Parameter(
        name="compressed_reserves_pre_swap",
        shape=(num_pools, num_tokens),
        value=np.array(
            (
                _compressed_starting_reserves_pool_hi,
                _compressed_starting_reserves_pool_lo,
            ),
            dtype=np.float64,
        ),
    )

    pool_hi_pre_swap_k = cvxpy.Parameter(
        name="pool_hi_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_hi_index]).value,
    )
    pool_lo_pre_swap_k = cvxpy.Parameter(
        name="pool_lo_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_lo_index]).value,
    )

    pool_lo_profit_token_in = cvxpy.Variable(
        name="pool_lo_profit_token_in",
        nonneg=True,
    )
    pool_hi_profit_token_out = cvxpy.Variable(
        name="pool_hi_profit_token_out",
        nonneg=True,
    )
    forward_token_amount = cvxpy.Variable(
        name="forward_token_amount",
        nonneg=True,
    )

    pool_hi_deposits = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    pool_lo_deposits = (
        (0, pool_lo_profit_token_in) if forward_token_index == 0 else (pool_lo_profit_token_in, 0)
    )
    deposits = cvxpy_bmat(
        (
            pool_hi_deposits,
            pool_lo_deposits,
        )
    )

    pool_hi_withdrawals = (
        (0, pool_hi_profit_token_out) if forward_token_index == 0 else (pool_hi_profit_token_out, 0)
    )
    pool_lo_withdrawals = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    withdrawals = cvxpy_bmat(
        (
            pool_hi_withdrawals,
            pool_lo_withdrawals,
        )
    )

    fees_removed = cvxpy_multiply(fee_multiplier, deposits)

    compressed_reserves_post_swap = (
        compressed_reserves_pre_swap + deposits - withdrawals - fees_removed
    )

    final_reserves = compressed_reserves_post_swap + fees_removed

    pool_hi_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_hi_index])
    pool_lo_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_lo_index])

    pool_hi_final_k = geo_mean(final_reserves[pool_hi_index])
    pool_lo_final_k = geo_mean(final_reserves[pool_lo_index])

    objective = cvxpy.Maximize(cvxpy_sum((withdrawals - deposits)[:, profit_token_index]))
    constraints = [
        # Pool invariant (x*y=k)
        pool_hi_post_swap_k >= pool_hi_pre_swap_k,
        pool_lo_post_swap_k >= pool_lo_pre_swap_k,
        # Withdrawals can't exceed pool reserves
        pool_hi_profit_token_out <= compressed_reserves_pre_swap[pool_hi_index, profit_token_index],
        forward_token_amount <= compressed_reserves_pre_swap[pool_lo_index, forward_token_index],
    ]

    problem = cvxpy.Problem(objective, constraints)
    problem.solve(
        solver=cvxpy.CLARABEL,
        # verbose=True,
    )

    assert problem.status in cvxpy.settings.SOLUTION_PRESENT

    uncompressed_forward_token_amount = min(
        int(
            cast("float", forward_token_amount.value)
            * compression_factor
            * 10**forward_token.decimals
        ),
        (
            pool_lo.state.reserves_token0
            if forward_token_index == 0
            else pool_lo.state.reserves_token1
        )
        - 1,
    )
    uncompressed_deposits = compression_factor * deposits
    uncompressed_withdrawals = compression_factor * withdrawals

    print()
    print("Solved")
    print(
        f"fee_multiplier                        = {[(float(fee[0]), float(fee[1])) for fee in fee_multiplier.value]}"
    )
    print(f"forward_token_amount                  = {uncompressed_forward_token_amount}")
    print(
        f"uncompressed withdrawals (pool_hi)    = {uncompressed_withdrawals[pool_hi_index].value}"
    )
    print(
        f"uncompressed withdrawals (pool_lo)    = {uncompressed_withdrawals[pool_lo_index].value}"
    )
    print(f"uncompressed deposits    (pool_hi)    = {uncompressed_deposits[pool_hi_index].value}")
    print(f"uncompressed deposits    (pool_lo)    = {uncompressed_deposits[pool_lo_index].value}")
    print(
        f"reserves_starting    (pool_hi)        = {[res for res in compressed_reserves_pre_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_ending      (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_final       (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_pre_swap    (pool_lo)        = {[res for res in compressed_reserves_pre_swap[pool_lo_index].value]}"
    )
    print(
        f"reserves_post_swap   (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}"
    )
    print(
        f"reserves_final       (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}"
    )
    print(f"pool_lo_pre_swap_k                    = {pool_lo_pre_swap_k.value}")
    print(f"pool_lo_post_swap_k                   = {pool_lo_post_swap_k.value}")
    print(f"pool_lo_final_k                       = {pool_lo_final_k.value}")
    print(f"pool_hi_pre_swap_k                    = {pool_hi_pre_swap_k.value}")
    print(f"pool_hi_post_swap_k                   = {pool_hi_post_swap_k.value}")
    print(f"pool_hi_final_k                       = {pool_hi_final_k.value}")

    # assert pool_hi_final_k.value >= pool_hi_start_k.value
    # assert pool_lo_final_k.value >= pool_lo_start_k.value

    weth_out = pool_hi.calculate_tokens_out_from_tokens_in(
        token_in=forward_token,
        token_in_quantity=uncompressed_forward_token_amount,
    )

    weth_in = pool_lo.calculate_tokens_in_from_tokens_out(
        token_out=forward_token,
        token_out_quantity=uncompressed_forward_token_amount,
    )
    print(f"Actual profit                         = {weth_out - weth_in}")


def test_2pool_uniswap_v2_double_decimal_corrected(
    wbtc_pool_a: MockLiquidityPool,
    wbtc_pool_b: MockLiquidityPool,
    weth_token: Erc20Token,
):
    profit_token = weth_token
    pool_a_roe = wbtc_pool_a.get_absolute_exchange_rate(token=profit_token)
    pool_b_roe = wbtc_pool_b.get_absolute_exchange_rate(token=profit_token)

    if pool_a_roe > pool_b_roe:
        pool_hi = wbtc_pool_a
        pool_lo = wbtc_pool_b
    else:
        pool_hi = wbtc_pool_b
        pool_lo = wbtc_pool_a

    num_pools = 2
    num_tokens = 2

    # Indices are arbitrary but must be consistent so amounts are consistent between pools in the
    # various arrays
    pool_hi_index, pool_lo_index = 0, 1

    token0_decimals = pool_hi.token0.decimals
    token1_decimals = pool_hi.token1.decimals

    forward_token = pool_hi.token1 if pool_hi.token0 == profit_token else pool_hi.token0
    forward_token_index = 1 if pool_hi.token0 == profit_token else 0
    profit_token_index = 0 if pool_hi.token0 == profit_token else 1
    assert forward_token_index != profit_token_index

    pool_hi_fees = [pool_hi.fee_token0, pool_hi.fee_token1]
    pool_lo_fees = [pool_lo.fee_token0, pool_lo.fee_token1]
    fee_multiplier = cvxpy_bmat(
        (
            pool_hi_fees,
            pool_lo_fees,
        )
    )

    # Identify the largest value to use as a common divisor for each token.
    compression_factor_token0 = max(
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals),
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals),
    )
    compression_factor_token1 = max(
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals),
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals),
    )
    compression_factor_forward_token = (
        compression_factor_token0 if forward_token_index == 0 else compression_factor_token1
    )
    print(f"compression factor (token0): {float(compression_factor_token0)}")
    print(f"compression factor (token1): {float(compression_factor_token1)}")

    # Compress all pool reserves into a 0.0 - 1.0 value range
    _compressed_starting_reserves_pool_hi = (
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals) / compression_factor_token0,
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals) / compression_factor_token1,
    )
    _compressed_starting_reserves_pool_lo = (
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals) / compression_factor_token0,
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals) / compression_factor_token1,
    )
    compressed_reserves_pre_swap = cvxpy.Parameter(
        name="compressed_reserves_pre_swap",
        shape=(num_pools, num_tokens),
        value=np.array(
            (
                _compressed_starting_reserves_pool_hi,
                _compressed_starting_reserves_pool_lo,
            ),
            dtype=np.float64,
        ),
    )

    pool_hi_pre_swap_k = cvxpy.Parameter(
        name="pool_hi_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_hi_index]).value,
    )
    pool_lo_pre_swap_k = cvxpy.Parameter(
        name="pool_lo_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_lo_index]).value,
    )

    pool_lo_profit_token_in = cvxpy.Variable(
        name="pool_lo_profit_token_in",
        nonneg=True,
    )
    pool_hi_profit_token_out = cvxpy.Variable(
        name="pool_hi_profit_token_out",
        nonneg=True,
    )
    forward_token_amount = cvxpy.Variable(
        name="forward_token_amount",
        nonneg=True,
    )

    pool_hi_deposits = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    pool_lo_deposits = (
        (0, pool_lo_profit_token_in) if forward_token_index == 0 else (pool_lo_profit_token_in, 0)
    )
    deposits = cvxpy_bmat(
        (
            pool_hi_deposits,
            pool_lo_deposits,
        )
    )

    pool_hi_withdrawals = (
        (0, pool_hi_profit_token_out) if forward_token_index == 0 else (pool_hi_profit_token_out, 0)
    )
    pool_lo_withdrawals = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    withdrawals = cvxpy_bmat(
        (
            pool_hi_withdrawals,
            pool_lo_withdrawals,
        )
    )

    fees_removed = cvxpy_multiply(fee_multiplier, deposits)

    compressed_reserves_post_swap = (
        compressed_reserves_pre_swap + deposits - withdrawals - fees_removed
    )

    final_reserves = compressed_reserves_post_swap + fees_removed

    pool_hi_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_hi_index])
    pool_lo_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_lo_index])

    pool_hi_final_k = geo_mean(final_reserves[pool_hi_index])
    pool_lo_final_k = geo_mean(final_reserves[pool_lo_index])

    objective = cvxpy.Maximize(cvxpy_sum((withdrawals - deposits)[:, profit_token_index]))
    constraints = [
        # Pool invariant (x*y=k)
        pool_hi_post_swap_k >= pool_hi_pre_swap_k,
        pool_lo_post_swap_k >= pool_lo_pre_swap_k,
        # Withdrawals can't exceed pool reserves
        pool_hi_profit_token_out <= compressed_reserves_pre_swap[pool_hi_index, profit_token_index],
        forward_token_amount <= compressed_reserves_pre_swap[pool_lo_index, forward_token_index],
    ]

    problem = cvxpy.Problem(objective, constraints)
    problem.solve(
        solver=cvxpy.CLARABEL,
        # verbose=True,
    )

    assert problem.status in cvxpy.settings.SOLUTION_PRESENT

    uncompressed_forward_token_amount = min(
        int(
            cast("float", forward_token_amount.value)
            * compression_factor_forward_token
            * 10**forward_token.decimals
        ),
        (
            pool_lo.state.reserves_token0
            if forward_token_index == 0
            else pool_lo.state.reserves_token1
        )
        - 1,
    )
    uncompressed_deposits = cvxpy_multiply(
        deposits, np.array([compression_factor_token0, compression_factor_token1])
    )
    uncompressed_withdrawals = cvxpy_multiply(
        withdrawals, np.array([compression_factor_token1, compression_factor_token1])
    )
    # uncompressed_deposits = compression_factor * deposits
    # uncompressed_withdrawals = compression_factor * withdrawals

    print()
    print("Solved")
    print(
        f"fee_multiplier                        = {[(float(fee[0]), float(fee[1])) for fee in fee_multiplier.value]}"
    )
    print(f"forward_token_amount                  = {uncompressed_forward_token_amount}")
    print(
        f"uncompressed withdrawals (pool_hi)    = {uncompressed_withdrawals[pool_hi_index].value}"
    )
    print(
        f"uncompressed withdrawals (pool_lo)    = {uncompressed_withdrawals[pool_lo_index].value}"
    )
    print(f"uncompressed deposits    (pool_hi)    = {uncompressed_deposits[pool_hi_index].value}")
    print(f"uncompressed deposits    (pool_lo)    = {uncompressed_deposits[pool_lo_index].value}")
    print(
        f"reserves_starting    (pool_hi)        = {[res for res in compressed_reserves_pre_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_ending      (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_final       (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}"
    )
    print(
        f"reserves_pre_swap    (pool_lo)        = {[res for res in compressed_reserves_pre_swap[pool_lo_index].value]}"
    )
    print(
        f"reserves_post_swap   (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}"
    )
    print(
        f"reserves_final       (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}"
    )
    print(f"pool_lo_pre_swap_k                    = {pool_lo_pre_swap_k.value}")
    print(f"pool_lo_post_swap_k                   = {pool_lo_post_swap_k.value}")
    print(f"pool_lo_final_k                       = {pool_lo_final_k.value}")
    print(f"pool_hi_pre_swap_k                    = {pool_hi_pre_swap_k.value}")
    print(f"pool_hi_post_swap_k                   = {pool_hi_post_swap_k.value}")
    print(f"pool_hi_final_k                       = {pool_hi_final_k.value}")

    # assert pool_hi_final_k.value >= pool_hi_start_k.value
    # assert pool_lo_final_k.value >= pool_lo_start_k.value

    weth_out = pool_hi.calculate_tokens_out_from_tokens_in(
        token_in=forward_token,
        token_in_quantity=uncompressed_forward_token_amount,
    )

    weth_in = pool_lo.calculate_tokens_in_from_tokens_out(
        token_out=forward_token,
        token_out_quantity=uncompressed_forward_token_amount,
    )
    print(f"Actual profit                         = {weth_out - weth_in}")


def test_base_2pool(
    test_pool_a: MockLiquidityPool,
    test_pool_b: MockLiquidityPool,
    weth_base_token: Erc20Token,
):
    # scipy forward amount: 4323768730401916416
    # cvxpy forward amount: 4323770602979739649

    profit_token = weth_base_token

    # profit_token = weth_base_token
    pool_a_roe = test_pool_a.get_absolute_exchange_rate(token=profit_token)
    pool_b_roe = test_pool_b.get_absolute_exchange_rate(token=profit_token)

    if pool_a_roe > pool_b_roe:
        pool_hi = test_pool_a
        pool_lo = test_pool_b
    else:
        pool_hi = test_pool_b
        pool_lo = test_pool_a

    assert pool_hi == test_pool_a
    assert pool_lo == test_pool_b

    num_pools = 2
    num_tokens = 2

    # Indices are arbitrary but must be consistent so amounts are consistent between pools in the
    # various arrays
    pool_hi_index, pool_lo_index = 0, 1

    token0_decimals = pool_hi.token0.decimals
    token1_decimals = pool_hi.token1.decimals

    forward_token = pool_hi.token1 if pool_hi.token0 == profit_token else pool_hi.token0
    forward_token_index = 1 if pool_hi.token0 == profit_token else 0
    profit_token_index = 0 if pool_hi.token0 == profit_token else 1
    assert forward_token_index != profit_token_index

    profit_token_decimals = token0_decimals if profit_token_index == 0 else token1_decimals

    pool_hi_fees = [pool_hi.fee_token0, pool_hi.fee_token1]
    pool_lo_fees = [pool_lo.fee_token0, pool_lo.fee_token1]
    fee_multiplier = cvxpy_bmat(
        (
            pool_hi_fees,
            pool_lo_fees,
        )
    )

    compression_factor_token0 = max(
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals),
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals),
    )
    compression_factor_token1 = max(
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals),
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals),
    )
    compression_factor_forward_token = (
        compression_factor_token0 if forward_token_index == 0 else compression_factor_token1
    )
    compression_factor_profit_token = (
        compression_factor_token0 if profit_token_index == 0 else compression_factor_token1
    )

    # Compress all pool reserves into a 0.0 - 1.0 value range
    _compressed_starting_reserves_pool_hi = (
        Fraction(pool_hi.state.reserves_token0, 10**token0_decimals) / compression_factor_token0,
        Fraction(pool_hi.state.reserves_token1, 10**token1_decimals) / compression_factor_token1,
    )
    _compressed_starting_reserves_pool_lo = (
        Fraction(pool_lo.state.reserves_token0, 10**token0_decimals) / compression_factor_token0,
        Fraction(pool_lo.state.reserves_token1, 10**token1_decimals) / compression_factor_token1,
    )
    compressed_reserves_pre_swap = cvxpy.Parameter(
        name="compressed_reserves_pre_swap",
        shape=(num_pools, num_tokens),
        value=np.array(
            (
                _compressed_starting_reserves_pool_hi,
                _compressed_starting_reserves_pool_lo,
            ),
            dtype=np.float64,
        ),
    )

    pool_hi_pre_swap_k = cvxpy.Parameter(
        name="pool_hi_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_hi_index]).value,
    )
    pool_lo_pre_swap_k = cvxpy.Parameter(
        name="pool_lo_pre_swap_k",
        value=geo_mean(compressed_reserves_pre_swap[pool_lo_index]).value,
    )

    pool_lo_profit_token_in = cvxpy.Variable(
        name="pool_lo_profit_token_in",
        nonneg=True,
    )
    pool_hi_profit_token_out = cvxpy.Variable(
        name="pool_hi_profit_token_out",
        nonneg=True,
    )
    forward_token_amount = cvxpy.Variable(
        name="forward_token_amount",
        nonneg=True,
    )

    pool_hi_deposits = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    pool_lo_deposits = (
        (0, pool_lo_profit_token_in) if forward_token_index == 0 else (pool_lo_profit_token_in, 0)
    )
    deposits = cvxpy_bmat(
        (
            pool_hi_deposits,
            pool_lo_deposits,
        )
    )

    pool_hi_withdrawals = (
        (0, pool_hi_profit_token_out) if forward_token_index == 0 else (pool_hi_profit_token_out, 0)
    )
    pool_lo_withdrawals = (
        (forward_token_amount, 0) if forward_token_index == 0 else (0, forward_token_amount)
    )
    withdrawals = cvxpy_bmat(
        (
            pool_hi_withdrawals,
            pool_lo_withdrawals,
        )
    )

    fees_removed = cvxpy_multiply(fee_multiplier, deposits)

    compressed_reserves_post_swap = (
        compressed_reserves_pre_swap + deposits - withdrawals - fees_removed
    )

    final_reserves = compressed_reserves_post_swap + fees_removed

    pool_hi_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_hi_index])
    pool_lo_post_swap_k = geo_mean(compressed_reserves_post_swap[pool_lo_index])

    pool_hi_final_k = geo_mean(final_reserves[pool_hi_index])
    pool_lo_final_k = geo_mean(final_reserves[pool_lo_index])

    objective = cvxpy.Maximize(pool_hi_profit_token_out - pool_lo_profit_token_in)
    constraints = [
        # Pool invariant (x*y=k)
        pool_hi_post_swap_k >= pool_hi_pre_swap_k,
        pool_lo_post_swap_k >= pool_lo_pre_swap_k,
        # Withdrawals can't exceed pool reserves
        # pool_hi_profit_token_out <= compressed_reserves_pre_swap[pool_hi_index, profit_token_index],
        # forward_token_amount <= compressed_reserves_pre_swap[pool_lo_index, forward_token_index],
    ]

    problem = cvxpy.Problem(objective, constraints)
    # clarabel_tols = 1e-8
    problem.solve(
        solver=cvxpy.CLARABEL,
        verbose=True,
        # tol_gap_abs=clarabel_tols,  # absolute duality gap tolerance
        # tol_gap_rel=clarabel_tols,  # relative duality gap tolerance
        # tol_feas=clarabel_tols,  # feasibility check tolerance (primal and dual)
        # tol_infeas_abs=clarabel_tols,  # absolute infeasibility tolerance (primal and dual)
        # tol_infeas_rel=clarabel_tols,  # relative infeasibility tolerance (primal and dual)
        # static_regularization_enable=False,
        # dynamic_regularization_enable=False,
        # iterative_refinement_enable=False,
        # equilibrate_enable=False,
    )

    assert problem.status in cvxpy.settings.SOLUTION_PRESENT

    uncompressed_forward_token_amount = min(
        int(
            cast("float", forward_token_amount.value)
            * compression_factor_forward_token
            * 10**forward_token.decimals
        ),
        (
            pool_lo.state.reserves_token0
            if forward_token_index == 0
            else pool_lo.state.reserves_token1
        )
        - 1,
    )
    uncompressed_deposits = cvxpy_multiply(
        deposits, np.array([compression_factor_token0, compression_factor_token1])
    )
    uncompressed_withdrawals = cvxpy_multiply(
        withdrawals, np.array([compression_factor_token0, compression_factor_token1])
    )
    # fmt:off
    print()
    print("Solved")
    print(f"fee_multiplier                        = {[(float(fee[0]), float(fee[1])) for fee in fee_multiplier.value]}")
    print(f"forward_token_amount                  = {uncompressed_forward_token_amount}")
    print(f"uncompressed withdrawals (pool_hi)    = {uncompressed_withdrawals[pool_hi_index].value}")
    print(f"uncompressed withdrawals (pool_lo)    = {uncompressed_withdrawals[pool_lo_index].value}")
    print(f"uncompressed deposits    (pool_hi)    = {uncompressed_deposits[pool_hi_index].value}")
    print(f"uncompressed deposits    (pool_lo)    = {uncompressed_deposits[pool_lo_index].value}")
    print(f"compressed withdrawals   (pool_hi)    = {withdrawals[pool_hi_index].value}")
    print(f"compressed withdrawals   (pool_lo)    = {withdrawals[pool_lo_index].value}")
    print(f"compressed deposits      (pool_hi)    = {deposits[pool_hi_index].value}")
    print(f"compressed deposits      (pool_lo)    = {deposits[pool_lo_index].value}")
    print(f"reserves_pre_swap    (pool_hi)        = {[res for res in compressed_reserves_pre_swap[pool_hi_index].value]}")
    print(f"reserves_ending      (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}")
    print(f"reserves_final       (pool_hi)        = {[res for res in compressed_reserves_post_swap[pool_hi_index].value]}")
    print(f"reserves_pre_swap    (pool_lo)        = {[res for res in compressed_reserves_pre_swap[pool_lo_index].value]}")
    print(f"reserves_post_swap   (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}")
    print(f"reserves_final       (pool_lo)        = {[res for res in compressed_reserves_post_swap[pool_lo_index].value]}")
    print(f"pool_lo_pre_swap_k                    = {pool_lo_pre_swap_k.value}")
    print(f"pool_lo_post_swap_k                   = {pool_lo_post_swap_k.value}")
    print(f"pool_lo_final_k                       = {pool_lo_final_k.value}")
    print(f"pool_hi_pre_swap_k                    = {pool_hi_pre_swap_k.value}")
    print(f"pool_hi_post_swap_k                   = {pool_hi_post_swap_k.value}")
    print(f"pool_hi_final_k                       = {pool_hi_final_k.value}")
    # fmt:on

    # assert pool_hi_final_k.value >= pool_hi_start_k.value
    # assert pool_lo_final_k.value >= pool_lo_start_k.value

    weth_out = pool_hi.calculate_tokens_out_from_tokens_in(
        token_in=forward_token,
        token_in_quantity=uncompressed_forward_token_amount,
    )

    weth_in = pool_lo.calculate_tokens_in_from_tokens_out(
        token_out=forward_token,
        token_out_quantity=uncompressed_forward_token_amount,
    )
    # fmt: off
    print(f"Fees (pool_hi)                        = {fees_removed[pool_hi_index, :].value}")
    print(f"Fees (pool_lo)                        = {fees_removed[pool_lo_index, :].value}")
    print(f"WETH out (approximate)                = {int(compression_factor_profit_token * 10**profit_token_decimals * pool_hi_profit_token_out.value)}")
    print(f"WETH out (actual)                     = {weth_out}")
    print(f"WETH in  (approximate)                = {int(compression_factor_profit_token * 10**profit_token_decimals * pool_lo_profit_token_in.value)}")
    print(f"WETH in  (actual)                     = {weth_in}")
    print(f"Profit (approximate)                  = {int(compression_factor_profit_token * 10**profit_token_decimals * pool_hi_profit_token_out.value) - int(compression_factor_profit_token * 10**profit_token_decimals * pool_lo_profit_token_in.value)}")
    print(f"Profit (actual)                       = {weth_out - weth_in}")
    # fmt:on


def test_base_3pool(
    link_token: Erc20Token,
    wbtc_token: Erc20Token,
    weth_token: Erc20Token,
):
    # fake prices:
    # 1 WTBC = 20 WETH
    # 1 WETH = 200 LINK
    # 1 WBTC = 4000 LINK

    lp_1 = MockLiquidityPool()
    lp_1.name = "Fake V2 Pool: LINK-WETH (0.3%)"
    lp_1.address = get_checksum_address("0x0000000000000000000000000000000000000001")
    lp_1.factory = get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")
    lp_1.fee_token0 = Fraction(3, 1000)
    lp_1.fee_token1 = Fraction(3, 1000)
    lp_1.token0 = link_token
    lp_1.token1 = weth_token
    lp_1._state = UniswapV2PoolState(
        address=lp_1.address,
        block=None,
        reserves_token0=20_000 * 10**link_token.decimals,
        reserves_token1=100 * 10**weth_token.decimals,
        # total liquidity 10_000 WETH
    )

    lp_2 = MockLiquidityPool()
    lp_2.name = "Fake V2 Pool: LINK-WBTC (0.3%)"
    lp_2.address = get_checksum_address("0x0000000000000000000000000000000000000002")
    lp_2.factory = get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")
    lp_2.fee_token0 = Fraction(3, 1000)
    lp_2.fee_token1 = Fraction(3, 1000)
    lp_2.token0 = wbtc_token
    lp_2.token1 = link_token
    lp_2._state = UniswapV2PoolState(
        address=lp_2.address,
        block=None,
        # reserves_token0=5 * 10**wbtc_token.decimals, # original balance
        reserves_token0=(
            # dislocate middle pool by decreasing the price of WBTC
            10 * 10**wbtc_token.decimals
        ),
        reserves_token1=20_000 * 10**link_token.decimals,
    )

    lp_3 = MockLiquidityPool()
    lp_3.name = "Fake V2 Pool: WBTC-WETH (0.3%)"
    lp_3.address = get_checksum_address("0x0000000000000000000000000000000000000003")
    lp_3.factory = get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f")
    lp_3.fee_token0 = Fraction(3, 1000)
    lp_3.fee_token1 = Fraction(3, 1000)
    lp_3.token0 = wbtc_token
    lp_3.token1 = weth_token
    lp_3._state = UniswapV2PoolState(
        address=lp_3.address,
        block=None,
        reserves_token0=5 * 10**wbtc_token.decimals,
        reserves_token1=100 * 10**weth_token.decimals,
        # total liquidity 10_000 WETH
    )

    arb = _UniswapMultiPoolCycleTesting(
        input_token=weth_token,
        # swap_pools=[wbtc_pool_b, wbtc_pool_a],
        swap_pools=[lp_1, lp_2, lp_3],  # profitable
        # swap_pools=[lp_3, lp_2, lp_1], # unprofitable
        id="test",
    )

    result = arb.calculate()
    print(f"{result.profit_amount=}")
    # print(f"{result=}")
    # payloads = arb.generate_payloads(
    #     from_address="0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f",
    #     pool_swap_amounts=result.swap_amounts,
    # )



from dataclasses import dataclass

import eth_typing
from eth_typing import ChecksumAddress

from degenbot import get_checksum_address
from degenbot.exceptions import DegenbotValueError
from degenbot.types.abstract import AbstractExchangeDeployment
from degenbot.types.aliases import ChainId


@dataclass(slots=True, frozen=True)
class UniswapFactoryDeployment:
    address: ChecksumAddress
    deployer: ChecksumAddress | None
    pool_init_hash: str


@dataclass(slots=True, frozen=True)
class UniswapV2ExchangeDeployment(AbstractExchangeDeployment):
    factory: UniswapFactoryDeployment


@dataclass(slots=True, frozen=True)
class UniswapV3ExchangeDeployment(AbstractExchangeDeployment):
    factory: UniswapFactoryDeployment


@dataclass(slots=True, frozen=True)
class UniswapRouterDeployment:
    address: ChecksumAddress
    chain_id: ChainId
    name: str
    exchanges: list[UniswapV2ExchangeDeployment | UniswapV3ExchangeDeployment]


def register_exchange(exchange: UniswapV2ExchangeDeployment | UniswapV3ExchangeDeployment) -> None:
    if exchange.chain_id not in FACTORY_DEPLOYMENTS:
        FACTORY_DEPLOYMENTS[exchange.chain_id] = {}

    if exchange.factory.address in FACTORY_DEPLOYMENTS[exchange.chain_id]:
        raise DegenbotValueError(message="Exchange is already registered.")

    FACTORY_DEPLOYMENTS[exchange.chain_id][exchange.factory.address] = exchange.factory


def register_router(router: UniswapRouterDeployment) -> None:
    if router.chain_id not in ROUTER_DEPLOYMENTS:
        ROUTER_DEPLOYMENTS[router.chain_id] = {}

    if router.address in ROUTER_DEPLOYMENTS[router.chain_id]:
        raise DegenbotValueError(message="Router is already registered.")

    ROUTER_DEPLOYMENTS[router.chain_id][router.address] = router


# Mainnet DEX --------------- START
EthereumMainnetUniswapV2 = UniswapV2ExchangeDeployment(
    name="Ethereum Mainnet Uniswap V2",
    chain_id=eth_typing.ChainId.ETH,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f"),
        deployer=None,
        pool_init_hash="0x96e8ac4277198ff8b6f785478aa9a39f403cb768dd02cbee326c3e7da348845f",
    ),
)
EthereumMainnetSushiswapV2 = UniswapV2ExchangeDeployment(
    name="Ethereum Mainnet Sushiswap V2",
    chain_id=eth_typing.ChainId.ETH,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"),
        deployer=None,
        pool_init_hash="0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303",
    ),
)
EthereumMainnetUniswapV3 = UniswapV3ExchangeDeployment(
    name="Ethereum Mainnet Uniswap V3",
    chain_id=eth_typing.ChainId.ETH,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
EthereumMainnetSushiswapV3 = UniswapV3ExchangeDeployment(
    name="Ethereum Mainnet Sushiswap V3",
    chain_id=eth_typing.ChainId.ETH,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0xbACEB8eC6b9355Dfc0269C18bac9d6E2Bdc29C4F"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
# Mainnet DEX --------------- END


# Base DEX ---------------- START
BaseAerodromeV2 = UniswapV2ExchangeDeployment(
    name="Aerodrome V2",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x420DD381b31aEf6683db6B902084cB0FFECe40Da"),
        deployer=None,
        pool_init_hash="",
    ),
)
BaseAerodromeV3 = UniswapV3ExchangeDeployment(
    name="Base Aerodrome V3",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x5e7BB104d84c7CB9B682AaC2F3d509f5F406809A"),
        deployer=None,
        pool_init_hash="",
    ),
)
BasePancakeswapV2 = UniswapV2ExchangeDeployment(
    name="Pancakeswap V2",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x02a84c1b3BBD7401a5f7fa98a384EBC70bB5749E"),
        deployer=None,
        pool_init_hash="0x57224589c67f3f30a6b0d7a1b54cf3153ab84563bc609ef41dfb34f8b2974d2d",
    ),
)
BasePancakeswapV3 = UniswapV3ExchangeDeployment(
    name="Pancakeswap V3",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x0BFbCF9fa4f9C56B0F40a671Ad40E0805A091865"),
        deployer=get_checksum_address("0x41ff9AA7e16B8B1a8a8dc4f0eFacd93D02d071c9"),
        pool_init_hash="0x6ce8eb472fa82df5469c6ab6d485f17c3ad13c8cd7af59b3d4a8026c5ce0f7e2",
    ),
)
BaseSushiswapV2 = UniswapV2ExchangeDeployment(
    name="Sushiswap V2",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x71524B4f93c58fcbF659783284E38825f0622859"),
        deployer=None,
        pool_init_hash="0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303",
    ),
)
BaseSushiswapV3 = UniswapV3ExchangeDeployment(
    name="Sushiswap V3",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0xc35DADB65012eC5796536bD9864eD8773aBc74C4"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
BaseSwapbasedV2 = UniswapV2ExchangeDeployment(
    name="Swapbased V2",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x04C9f118d21e8B767D2e50C946f0cC9F6C367300"),
        deployer=None,
        pool_init_hash="0xb64118b4e99d4a4163453838112a1695032df46c09f7f09064d4777d2767f8ea",
    ),
)
BaseUniswapV2 = UniswapV2ExchangeDeployment(
    name="Uniswap V2",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x8909Dc15e40173Ff4699343b6eB8132c65e18eC6"),
        deployer=None,
        pool_init_hash="0x96e8ac4277198ff8b6f785478aa9a39f403cb768dd02cbee326c3e7da348845f",
    ),
)
BaseUniswapV3 = UniswapV3ExchangeDeployment(
    name="Uniswap V3",
    chain_id=eth_typing.ChainId.BASE,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x33128a8fC17869897dcE68Ed026d694621f6FDfD"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
# Base DEX -------------------- END


# Arbitrum DEX -------------- START
ArbitrumSushiswapV2 = UniswapV2ExchangeDeployment(
    name="Arbitrum Sushiswap V2",
    chain_id=eth_typing.ChainId.ARB1,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0xc35DADB65012eC5796536bD9864eD8773aBc74C4"),
        deployer=None,
        pool_init_hash="0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303",
    ),
)
ArbitrumUniswapV3 = UniswapV3ExchangeDeployment(
    name="Arbitrum Uniswap V3",
    chain_id=eth_typing.ChainId.ARB1,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
ArbitrumSushiswapV3 = UniswapV3ExchangeDeployment(
    name="Arbitrum Sushiswap V3",
    chain_id=eth_typing.ChainId.ARB1,
    factory=UniswapFactoryDeployment(
        address=get_checksum_address("0x1af415a1EbA07a4986a52B6f2e7dE7003D82231e"),
        deployer=None,
        pool_init_hash="0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54",
    ),
)
# ----------------------------- END


# Mainnet Routers ----------- START
EthereumMainnetUniswapV2Router = UniswapRouterDeployment(
    address=get_checksum_address("0xf164fC0Ec4E93095b804a4795bBe1e041497b92a"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet UniswapV2 Router",
    exchanges=[EthereumMainnetUniswapV2],
)
EthereumMainnetUniswapV2Router2 = UniswapRouterDeployment(
    address=get_checksum_address("0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet UniswapV2 Router 2",
    exchanges=[EthereumMainnetUniswapV2],
)
EthereumMainnetSushiswapV2Router = UniswapRouterDeployment(
    address=get_checksum_address("0xd9e1cE17f2641f24aE83637ab66a2cca9C378B9F"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Sushiswap Router",
    exchanges=[EthereumMainnetSushiswapV2],
)
EthereumMainnetUniswapV3Router = UniswapRouterDeployment(
    address=get_checksum_address("0xE592427A0AEce92De3Edee1F18E0157C05861564"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Uniswap V3 Router",
    exchanges=[EthereumMainnetUniswapV3],
)
EthereumMainnetUniswapV3Router2 = UniswapRouterDeployment(
    address=get_checksum_address("0x68b3465833fb72A70ecDF485E0e4C7bD8665Fc45"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Uniswap V3 Router2",
    exchanges=[EthereumMainnetUniswapV2, EthereumMainnetUniswapV3],
)
EthereumMainnetUniswapUniversalRouter = UniswapRouterDeployment(
    address=get_checksum_address("0xEf1c6E67703c7BD7107eed8303Fbe6EC2554BF6B"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Uniswap Universal Router",
    exchanges=[EthereumMainnetUniswapV2, EthereumMainnetUniswapV3],
)
EthereumMainnetUniswapUniversalRouterV1_2 = UniswapRouterDeployment(
    address=get_checksum_address("0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Uniswap Universal Router V1_2",
    exchanges=[EthereumMainnetUniswapV2, EthereumMainnetUniswapV3],
)
EthereumMainnetUniswapUniversalRouterV1_3 = UniswapRouterDeployment(
    address=get_checksum_address("0x3F6328669a86bef431Dc6F9201A5B90F7975a023"),
    chain_id=eth_typing.ChainId.ETH,
    name="Ethereum Mainnet Uniswap Universal Router V1_3",
    exchanges=[EthereumMainnetUniswapV2, EthereumMainnetUniswapV3],
)
# Mainnet Routers ------------- END


# Base Routers -------------- START
BaseSushiswapRouter = UniswapRouterDeployment(
    address=get_checksum_address("0xFB7eF66a7e61224DD6FcD0D7d9C3be5C8B049b9f"),
    chain_id=eth_typing.ChainId.BASE,
    name="Sushiswap V3 SwapRouter",
    exchanges=[BaseSushiswapV3],
)
# Base Routers ---------------- END


# Arbitrum Routers ---------- START
ArbitrumSushiswapV2Router = UniswapRouterDeployment(
    address=get_checksum_address("0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Sushiswap Router",
    exchanges=[ArbitrumSushiswapV2],
)
ArbitrumUniswapV3Router = UniswapRouterDeployment(
    address=get_checksum_address("0xE592427A0AEce92De3Edee1F18E0157C05861564"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Uniswap V3 Router",
    exchanges=[ArbitrumUniswapV3],
)
ArbitrumUniswapV3Router2 = UniswapRouterDeployment(
    address=get_checksum_address("0x68b3465833fb72A70ecDF485E0e4C7bD8665Fc45"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Uniswap V3 Router2",
    exchanges=[ArbitrumUniswapV3],
)
ArbitrumUniswapUniversalRouter = UniswapRouterDeployment(
    address=get_checksum_address("0x4C60051384bd2d3C01bfc845Cf5F4b44bcbE9de5"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Uniswap Universal Router",
    exchanges=[ArbitrumUniswapV3],
)
ArbitrumUniswapUniversalRouter2 = UniswapRouterDeployment(
    address=get_checksum_address("0xeC8B0F7Ffe3ae75d7FfAb09429e3675bb63503e4"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Uniswap Universal Router V1_2",
    exchanges=[ArbitrumUniswapV3],
)
ArbitrumUniswapUniversalRouter3 = UniswapRouterDeployment(
    address=get_checksum_address("0x5E325eDA8064b456f4781070C0738d849c824258"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Uniswap Universal Router 4",
    exchanges=[ArbitrumUniswapV3],
)
ArbitrumSushiswapV3Router = UniswapRouterDeployment(
    address=get_checksum_address("0x8A21F6768C1f8075791D08546Dadf6daA0bE820c"),
    chain_id=eth_typing.ChainId.ARB1,
    name="Arbitrum Sushiswap V3 Router",
    exchanges=[ArbitrumSushiswapV3],
)
# ----------------------------- END


FACTORY_DEPLOYMENTS: dict[
    int,  # chain ID
    dict[ChecksumAddress, UniswapFactoryDeployment],
] = {
    eth_typing.ChainId.ETH: {
        EthereumMainnetSushiswapV2.factory.address: EthereumMainnetSushiswapV2.factory,
        EthereumMainnetSushiswapV3.factory.address: EthereumMainnetSushiswapV3.factory,
        EthereumMainnetUniswapV2.factory.address: EthereumMainnetUniswapV2.factory,
        EthereumMainnetUniswapV3.factory.address: EthereumMainnetUniswapV3.factory,
    },
    eth_typing.ChainId.BASE: {
        BaseAerodromeV2.factory.address: BaseAerodromeV2.factory,
        BaseAerodromeV3.factory.address: BaseAerodromeV3.factory,
        BasePancakeswapV2.factory.address: BasePancakeswapV2.factory,
        BasePancakeswapV3.factory.address: BasePancakeswapV3.factory,
        BaseSushiswapV2.factory.address: BaseSushiswapV2.factory,
        BaseSushiswapV3.factory.address: BaseSushiswapV3.factory,
        BaseSwapbasedV2.factory.address: BaseSwapbasedV2.factory,
        BaseUniswapV2.factory.address: BaseUniswapV2.factory,
        BaseUniswapV3.factory.address: BaseUniswapV3.factory,
    },
    eth_typing.ChainId.ARB1: {
        ArbitrumSushiswapV2.factory.address: ArbitrumSushiswapV2.factory,
        ArbitrumSushiswapV3.factory.address: ArbitrumSushiswapV3.factory,
        ArbitrumUniswapV3.factory.address: ArbitrumUniswapV3.factory,
    },
}


ROUTER_DEPLOYMENTS: dict[
    int,  # chain ID
    dict[
        ChecksumAddress,  # Router Address
        UniswapRouterDeployment,
    ],
] = {
    eth_typing.ChainId.ETH: {
        EthereumMainnetSushiswapV2Router.address: EthereumMainnetSushiswapV2Router,
        EthereumMainnetUniswapV2Router.address: EthereumMainnetUniswapV2Router,
        EthereumMainnetUniswapV2Router2.address: EthereumMainnetUniswapV2Router2,
        EthereumMainnetUniswapV3Router.address: EthereumMainnetUniswapV3Router,
        EthereumMainnetUniswapV3Router2.address: EthereumMainnetUniswapV3Router2,
        EthereumMainnetUniswapUniversalRouter.address: EthereumMainnetUniswapUniversalRouter,
        EthereumMainnetUniswapUniversalRouterV1_2.address: EthereumMainnetUniswapUniversalRouterV1_2,  # noqa: E501
        EthereumMainnetUniswapUniversalRouterV1_3.address: EthereumMainnetUniswapUniversalRouterV1_3,  # noqa: E501
    },
    eth_typing.ChainId.BASE: {
        BaseSushiswapRouter.address: BaseSushiswapRouter,
    },
    eth_typing.ChainId.ARB1: {
        ArbitrumUniswapUniversalRouter.address: ArbitrumUniswapUniversalRouter,
        ArbitrumUniswapUniversalRouter2.address: ArbitrumUniswapUniversalRouter2,
        ArbitrumUniswapUniversalRouter3.address: ArbitrumUniswapUniversalRouter3,
    },
}



import dataclasses

from degenbot.types.abstract import AbstractPoolState


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class BalancerV2PoolState(AbstractPoolState):
    balances: tuple[int, ...]



from degenbot.constants import MAX_UINT128
from degenbot.functions import evm_divide
from degenbot.uniswap.v3_libraries.tick_math import MAX_TICK, MIN_TICK


def tick_spacing_to_max_liquidity_per_tick(tick_spacing: int) -> int:
    min_tick = evm_divide(MIN_TICK, tick_spacing) * tick_spacing
    max_tick = (MAX_TICK // tick_spacing) * tick_spacing
    num_ticks = (max_tick - min_tick) // tick_spacing + 1
    return MAX_UINT128 // num_ticks



import dataclasses

from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from degenbot.erc20 import Erc20Token
from degenbot.types.aliases import BlockNumber


class AbstractSwapAmounts:
    pass


@dataclasses.dataclass(slots=True, frozen=True)
class ArbitrageCalculationResult:
    id: str
    input_token: Erc20Token
    profit_token: Erc20Token
    input_amount: int
    profit_amount: int
    swap_amounts: tuple[AbstractSwapAmounts, ...]
    state_block: BlockNumber | None

    def __post_init__(self) -> None:
        assert self.input_amount != 0


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolSwapAmounts(AbstractSwapAmounts):
    token_in: Erc20Token
    token_in_index: int
    token_out: Erc20Token
    token_out_index: int
    amount_in: int
    min_amount_out: int
    underlying: bool

    def __post_init__(self) -> None:
        assert self.token_in != self.token_out


@dataclasses.dataclass(slots=True)
class UniswapV2PoolSwapAmounts(AbstractSwapAmounts):
    pool: ChecksumAddress
    amounts_in: tuple[int, int]
    amounts_out: tuple[int, int]
    recipient: ChecksumAddress | None = None

    def __post_init__(self) -> None:
        assert self.amounts_in != (0, 0)
        assert self.amounts_out != (0, 0)
        assert 0 in self.amounts_in
        assert 0 in self.amounts_out


@dataclasses.dataclass(slots=True)
class UniswapV3PoolSwapAmounts(AbstractSwapAmounts):
    pool: ChecksumAddress
    amount_specified: int
    zero_for_one: bool
    sqrt_price_limit_x96: int
    recipient: ChecksumAddress | None = None

    def __post_init__(self) -> None:
        assert self.amount_specified != 0


@dataclasses.dataclass(slots=True)
class UniswapV4PoolSwapAmounts(AbstractSwapAmounts):
    address: ChecksumAddress
    id: HexBytes
    amount_specified: int
    zero_for_one: bool
    sqrt_price_limit_x96: int
    recipient: ChecksumAddress | None = None

    def __post_init__(self) -> None:
        assert self.amount_specified != 0


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolVector:
    token_in: Erc20Token
    token_out: Erc20Token

    def __post_init__(self) -> None:
        assert self.token_in != self.token_out



from degenbot.exceptions import DegenbotError

"""
Exceptions defined here are raised by classes and functions in the `registry` module.
"""


class RegistryError(DegenbotError):
    """
    Exception raised inside registries.
    """


class RegistryAlreadyInitialized(RegistryError):
    """
    Raised by a singleton registry if a caller attempts to recreate it.
    """




from pydantic import validate_call

from degenbot.validation.evm_values import ValidatedUint256, ValidatedUint256NonZero


@validate_call(validate_return=True)
def mulmod(
    x: ValidatedUint256,
    y: ValidatedUint256,
    k: ValidatedUint256NonZero,
) -> ValidatedUint256:
    """
    Returns (x*y)%k, as implemented by Yul.

    ref: https://docs.soliditylang.org/en/latest/yul.html
    """

    return 0 if k == 0 else (x * y) % k



from textwrap import dedent

import click
import tomlkit
from pydantic import TypeAdapter

from degenbot import __version__, settings
from degenbot.database import create_new_sqlite_database, upgrade_existing_sqlite_database


@click.group()
@click.version_option()
def cli() -> None: ...


@cli.group()
def config() -> None: ...


@config.command("show")
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    type=str,
    help="Show configuration in JSON format",
)
@click.option(
    "--toml",
    "output_format",
    flag_value="toml",
    type=str,
    help="Show configuration in TOML format (default)",
    default=True,
)
def config_show(output_format: str) -> None:
    match output_format:
        case "json":
            click.echo(
                TypeAdapter(dict).dump_json(
                    settings.model_dump(),
                    indent=2,
                ),
            )
        case "toml":
            click.echo(
                tomlkit.dumps(
                    settings.model_dump(),
                ),
            )
        case _:
            ...


@cli.group()
def database() -> None: ...


@database.command("reset")
def database_reset() -> None:
    user_confirm = click.confirm(
        dedent(
            f"""\
            The existing DB at {settings.database.path} will be removed and a new, empty DB will be created and initialized using the schema included in {__package__} version {__version__}.
            Do you want to proceed?"""  # noqa: E501
        ),
        default=False,
    )
    if user_confirm:
        create_new_sqlite_database(settings.database.path)
    else:
        raise click.Abort


@database.command("upgrade")
@click.option("--force", is_flag=True, help="Skip confirmation prompt")
def database_upgrade(*, force: bool) -> None:
    # TODO: convert placeholder values to real
    old = 1
    new = 2

    if force or click.confirm(
        dedent(
            f"""\
            The DB at {settings.database.path} will be upgraded from version {old} to {new}.
            Do you want to proceed?"""
        ),
        default=False,
    ):
        upgrade_existing_sqlite_database(settings.database.path)
    else:
        raise click.Abort


@database.command("verify")
def database_verify() -> None:
    click.echo(f"(placeholder) DB at {settings.database.path} verified")



from dataclasses import dataclass

from eth_typing import ChecksumAddress

from degenbot.types.abstract import AbstractExchangeDeployment, AbstractPoolState
from degenbot.types.concrete import PoolStateMessage
from degenbot.uniswap.v2_types import UniswapV2PoolExternalUpdate
from degenbot.uniswap.v3_types import UniswapV3PoolState


@dataclass(slots=True, frozen=True)
class SolidlyFactoryDeployment:
    address: ChecksumAddress
    deployer: ChecksumAddress | None
    pool_init_hash: str


@dataclass(slots=True, frozen=True)
class SolidlyExchangeDeployment(AbstractExchangeDeployment):
    factory: SolidlyFactoryDeployment


class AerodromeV2PoolExternalUpdate(UniswapV2PoolExternalUpdate): ...


@dataclass(slots=True, frozen=True)
class AerodromeV2PoolState(AbstractPoolState):
    reserves_token0: int
    reserves_token1: int


@dataclass(slots=True, frozen=True)
class AerodromeV2PoolSimulationResult:
    amount0_delta: int
    amount1_delta: int
    current_state: AerodromeV2PoolState
    future_state: AerodromeV2PoolState


@dataclass(slots=True, frozen=True)
class AerodromeV2PoolStateUpdated(PoolStateMessage):
    state: AerodromeV2PoolState


class AerodromeV3PoolState(UniswapV3PoolState): ...


@dataclass(slots=True, frozen=True)
class AerodromeV3PoolStateUpdated(PoolStateMessage):
    state: AerodromeV3PoolState



from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool


class SushiswapV2Pool(UniswapV2Pool): ...


class SushiswapV3Pool(UniswapV3Pool): ...



from eth_typing import ChecksumAddress

from degenbot.exceptions.evm import EVMRevertError


def ensure_input_length_match(*nums: int) -> None:
    if len(set(nums)) != 1:
        raise EVMRevertError(error="INPUT_LENGTH_MISMATCH")


def ensure_array_is_sorted(array: list[ChecksumAddress]) -> None:
    if len(array) <= 1:
        return

    if sorted(array) != array:
        raise EVMRevertError(error="UNSORTED_ARRAY")



from degenbot.functions import raise_if_invalid_uint256
from degenbot.solidly.solidly_functions import general_calc_d


def f_camelot(x0: int, y: int) -> int:
    return (
        x0 * (y * y // 10**18 * y // 10**18) // 10**18
        + (x0 * x0 // 10**18 * x0 // 10**18) * y // 10**18
    )


def get_y_camelot(
    x_0: int,
    xy: int,
    y: int,
) -> int:  # pragma: no cover
    for _ in range(255):
        y_prev = y
        k = f_camelot(x_0, y)
        if k < xy:
            dy = (xy - k) * 10**18 // general_calc_d(x_0, y)

            y = y + dy
        else:
            dy = (k - xy) * 10**18 // general_calc_d(x_0, y)
            y = y - dy

        if y > y_prev:
            if y - y_prev <= 1:
                return y
        elif y_prev - y <= 1:
            return y
    return y


def k_camelot(
    balance_0: int,
    balance_1: int,
    decimals_0: int,
    decimals_1: int,
) -> int:
    _x = balance_0 * 10**18 // decimals_0
    _y = balance_1 * 10**18 // decimals_1
    _a = (_x * _y) // 10**18
    _b = (_x * _x) // 10**18 + (_y * _y) // 10**18
    raise_if_invalid_uint256(_a * _b)
    return _a * _b // 10**18  # x^3*y + y^3*x >= k



from threading import Lock
from typing import Any, ClassVar, Self
from weakref import WeakValueDictionary

from eth_typing import ChecksumAddress

from degenbot import get_checksum_address
from degenbot.types.abstract.liquidity_pool import AbstractLiquidityPool
from degenbot.types.aliases import ChainId


class AbstractPoolManager[Pool: AbstractLiquidityPool]:
    """
    Base class for liquidity pool managers. The class instance dict and get_instance method are
    mechanisms for implementing a singleton strategy so only one pool manager is created for a given
    DEX factory.
    """

    # Class variables
    instances: ClassVar[
        WeakValueDictionary[
            tuple[ChainId, ChecksumAddress],
            Any,
        ]
    ] = WeakValueDictionary()
    pool_factory: type[Pool]  # this class attribute is set by __init_subclass__ at import time

    # Instance variables
    _chain_id: ChainId
    _deployer_address: ChecksumAddress
    _factory_address: ChecksumAddress
    _tracked_pools: dict[ChecksumAddress, Pool]
    _untracked_pools: set[ChecksumAddress]
    _lock: Lock

    def __init_subclass__(cls, *, pool_factory: type[Pool] | None = None, **kwargs: Any) -> None:
        if pool_factory is not None:
            cls.pool_factory = pool_factory
        super().__init_subclass__(**kwargs)

    @classmethod
    def get_instance(cls, factory_address: str, chain_id: ChainId) -> Self | None:
        return cls.instances.get((chain_id, get_checksum_address(factory_address)))

    def _add_tracked_pool(self, pool_helper: Pool) -> None:
        with self._lock:
            self._tracked_pools[pool_helper.address] = pool_helper

    def remove(self, pool_address: ChecksumAddress | str) -> None:
        pool_address = get_checksum_address(pool_address)
        self._tracked_pools.pop(pool_address, None)
        self._untracked_pools.discard(pool_address)

    @property
    def chain_id(self) -> ChainId:
        return self._chain_id

    @property
    def deployer_address(self) -> ChecksumAddress:
        return self._deployer_address

    @property
    def factory_address(self) -> ChecksumAddress:
        return self._factory_address



import contextlib
from typing import TYPE_CHECKING

from web3.types import BlockIdentifier

import degenbot.registry
from degenbot import get_checksum_address
from degenbot.connection import connection_manager
from degenbot.constants import ZERO_ADDRESS
from degenbot.erc20 import Erc20Token
from degenbot.functions import get_number_for_block_identifier
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import BoundedCache

if TYPE_CHECKING:
    from eth_typing import ChecksumAddress


class EtherPlaceholder(Erc20Token):
    """
    An Erc20Token-like adapter for pools using the 'all Es' or zero address placeholder to represent
    native Ether.
    """

    addresses = (
        ZERO_ADDRESS,
        get_checksum_address("0xEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE"),
    )
    symbol = "ETH"
    name = "Ether Placeholder"
    decimals = 18

    def __init__(
        self,
        address: str,
        *,
        chain_id: ChainId | None = None,
        state_cache_depth: int = 8,
    ) -> None:
        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id
        self._cached_balance: dict[ChecksumAddress, BoundedCache[BlockNumber, int]] = {}
        self.address = get_checksum_address(address)
        degenbot.registry.token_registry.add(
            token_address=self.address, chain_id=self._chain_id, token=self
        )
        self._state_cache_depth = state_cache_depth

    def get_balance(
        self,
        address: str,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        address = get_checksum_address(address)

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                self.w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_balance[address][block_number]

        balance = self.w3.eth.get_balance(
            address,
            block_identifier=block_number,
        )

        balance_cache_at_address: BoundedCache[BlockNumber, int]
        try:
            balance_cache_at_address = self._cached_balance[address]
        except KeyError:
            balance_cache_at_address = BoundedCache(max_items=self._state_cache_depth)

        balance_cache_at_address[block_number] = balance
        self._cached_balance[address] = balance_cache_at_address
        return balance



from collections.abc import Callable, Iterable, Iterator
from fractions import Fraction
from itertools import cycle
from typing import Final

import eth_abi.abi
from eth_typing import ChecksumAddress
from eth_utils.crypto import keccak
from hexbytes import HexBytes

from degenbot import get_checksum_address
from degenbot.exceptions import DegenbotValueError
from degenbot.functions import create2_address, evm_divide
from degenbot.uniswap.v3_libraries import tick_bitmap
from degenbot.uniswap.v3_types import Pip


def decode_v3_path(path: bytes) -> list[ChecksumAddress | Pip]:
    """
    Decode the `path` bytes used by the Uniswap V3 Router/Router2 contracts. `path` is a
    close-packed encoding of 20 byte pool addresses, interleaved with 3 byte fees.
    """
    address_bytes: Final = 20
    fee_bytes: Final = 3

    def _extract_address(chunk: bytes) -> ChecksumAddress:
        return get_checksum_address(chunk)

    def _extract_fee(chunk: bytes) -> Pip:
        return int.from_bytes(chunk, byteorder="big")

    if any(
        [
            len(path) < address_bytes + fee_bytes + address_bytes,
            len(path) % (address_bytes + fee_bytes) != address_bytes,
        ]
    ):  # pragma: no cover
        raise DegenbotValueError(message="Invalid path.")

    chunk_length_and_decoder_function: Iterator[
        tuple[
            int,
            Callable[
                [bytes],
                ChecksumAddress | Pip,
            ],
        ]
    ] = cycle(
        [
            (address_bytes, _extract_address),
            (fee_bytes, _extract_fee),
        ]
    )

    path_offset = 0
    decoded_path: list[ChecksumAddress | Pip] = []
    while path_offset != len(path):
        byte_length, extraction_func = next(chunk_length_and_decoder_function)
        path_chunk = HexBytes(path[path_offset : path_offset + byte_length])
        decoded_path.append(extraction_func(path_chunk))
        path_offset += byte_length

    return decoded_path


def exchange_rate_from_sqrt_price_x96(sqrt_price_x96: int) -> Fraction:
    # ref: https://blog.uniswap.org/uniswap-v3-math-primer

    # equivalent to Fraction(sqrt_price_x96, 2**96)**2
    return Fraction(
        sqrt_price_x96 * sqrt_price_x96,
        6277101735386680763835789423207666416102355444464034512896,  # 2**192
    )


def generate_v3_pool_address(
    deployer_address: str | bytes,
    token_addresses: Iterable[str | bytes],
    fee: Pip,
    init_hash: str | bytes,
) -> ChecksumAddress:
    """
    Get the deterministic V3 pool address generated by CREATE2. Uses the token address and fee to
    generate the salt. The token addresses can be passed in any order.

    Adapted from https://github.com/Uniswap/v3-periphery/blob/0682387198a24c7cd63566a2c58398533860a5d1/contracts/libraries/PoolAddress.sol#L33
    """

    token_addresses = sorted([HexBytes(address) for address in token_addresses])

    salt = keccak(
        eth_abi.abi.encode(
            ("address", "address", "uint24"),
            (*token_addresses, fee),
        )
    )

    return create2_address(
        deployer=deployer_address,
        salt=salt,
        init_code_hash=init_hash,
    )


def get_tick_word_and_bit_position(
    tick: int,
    tick_spacing: int,
) -> tuple[int, int]:
    """
    Retrieves the word and bit position for the tick, accounting for tick spacing.
    """
    return tick_bitmap.position(evm_divide(tick, tick_spacing))



from typing import TYPE_CHECKING

import pytest
import web3.middleware
from hexbytes import HexBytes
from pydantic import ValidationError

from degenbot import AnvilFork, get_checksum_address, set_web3
from degenbot.constants import MAX_UINT256, MIN_UINT256
from degenbot.exceptions import DegenbotValueError

from .conftest import (
    BASE_FULL_NODE_HTTP_URI,
    ETHEREUM_ARCHIVE_NODE_HTTP_URI,
    ETHEREUM_FULL_NODE_HTTP_URI,
)

if TYPE_CHECKING:
    from web3.providers.ipc import IPCProvider


VITALIK_ADDRESS = get_checksum_address("0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045")
WETH_ADDRESS = get_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")


def test_web3_endpoints():
    fork = AnvilFork(fork_url=ETHEREUM_FULL_NODE_HTTP_URI)
    assert fork.http_url == f"http://localhost:{fork.port}"
    assert fork.ws_url == f"ws://localhost:{fork.port}"

    current_block = fork.w3.eth.block_number
    assert web3.Web3(web3.HTTPProvider(fork.http_url)).eth.block_number == current_block
    assert web3.Web3(web3.LegacyWebSocketProvider(fork.ws_url)).eth.block_number == current_block


def test_set_bytecode():
    fake_bytecode = HexBytes("0x42069")
    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        bytecode_overrides=[
            (VITALIK_ADDRESS, fake_bytecode),
        ],
    )
    assert fork.w3.eth.get_code(VITALIK_ADDRESS) == fake_bytecode


def test_set_storage():
    storage_position = 0
    new_storage_value = HexBytes("0x42069")
    new_storage_value_padded = new_storage_value.hex().zfill(64)

    fork = AnvilFork(fork_url=ETHEREUM_FULL_NODE_HTTP_URI)
    assert fork.w3.eth.get_storage_at(
        account=WETH_ADDRESS,
        position=storage_position,
    ) != HexBytes(new_storage_value_padded)
    fork.set_storage(WETH_ADDRESS, position=storage_position, value=new_storage_value)

    assert fork.w3.eth.get_storage_at(
        account=WETH_ADDRESS,
        position=storage_position,
    ) == HexBytes(new_storage_value_padded)

    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        storage_overrides=[(WETH_ADDRESS, storage_position, new_storage_value)],
    )
    assert fork.w3.eth.get_storage_at(
        account=WETH_ADDRESS,
        position=storage_position,
    ) == HexBytes(new_storage_value_padded)


def test_rpc_methods(fork_mainnet_full: AnvilFork):
    with pytest.raises(ValidationError):
        fork_mainnet_full.set_next_base_fee(MIN_UINT256 - 1)
    with pytest.raises(ValidationError):
        fork_mainnet_full.set_next_base_fee(MAX_UINT256 + 1)
    fork_mainnet_full.set_next_base_fee(11 * 10**9)

    # Set several snapshot IDs and return to them
    snapshot_ids = [fork_mainnet_full.set_snapshot() for _ in range(10)]
    for snapshot_id in snapshot_ids:
        assert fork_mainnet_full.return_to_snapshot(snapshot_id) is True
    # No snapshot ID with this value
    assert fork_mainnet_full.return_to_snapshot(100) is False

    # Negative IDs are not allowed
    with pytest.raises(DegenbotValueError, match="ID cannot be negative"):
        fork_mainnet_full.return_to_snapshot(-1)

    for balance in [MIN_UINT256, MAX_UINT256]:
        fork_mainnet_full.set_balance(VITALIK_ADDRESS, balance)
        assert fork_mainnet_full.w3.eth.get_balance(VITALIK_ADDRESS) == balance

    # Balances outside of uint256 should be rejected
    with pytest.raises(ValidationError):
        fork_mainnet_full.set_balance(VITALIK_ADDRESS, MIN_UINT256 - 1)
    with pytest.raises(ValidationError):
        fork_mainnet_full.set_balance(VITALIK_ADDRESS, MAX_UINT256 + 1)

    fake_coinbase = get_checksum_address("0x0420042004200420042004200420042004200420")
    fork_mainnet_full.set_coinbase(fake_coinbase)
    # @dev the eth_coinbase method fails when called on Anvil,
    # so check by mining a block and comparing the miner address

    fork_mainnet_full.mine()
    block = fork_mainnet_full.w3.eth.get_block("latest")
    assert block.get("miner") == fake_coinbase


def test_mine_and_reset(fork_mainnet_full: AnvilFork):
    starting_block = fork_mainnet_full.w3.eth.get_block_number()
    fork_mainnet_full.mine()
    fork_mainnet_full.mine()
    fork_mainnet_full.mine()
    assert fork_mainnet_full.w3.eth.get_block_number() == starting_block + 3
    fork_mainnet_full.reset(block_number=starting_block)
    assert fork_mainnet_full.w3.eth.get_block_number() == starting_block


def test_fork_from_transaction_hash():
    fork = AnvilFork(
        fork_url=ETHEREUM_ARCHIVE_NODE_HTTP_URI,
        fork_transaction_hash="0x12167fa2a4cd676a6e740edb09427469ecb8718d84ef4d0d5819fe8b527964d6",
    )
    assert fork.w3.eth.block_number == 20987963


def test_set_next_block_base_fee(fork_mainnet_full: AnvilFork):
    base_fee_override = 69 * 10**9

    fork_mainnet_full.set_next_base_fee(base_fee_override)
    fork_mainnet_full.mine()
    assert fork_mainnet_full.w3.eth.get_block("latest")["baseFeePerGas"] == base_fee_override


def test_set_next_block_base_fee_in_constructor():
    base_fee_override = 69 * 10**9

    fork = AnvilFork(fork_url=ETHEREUM_FULL_NODE_HTTP_URI, base_fee=base_fee_override)
    fork.mine()
    assert fork.w3.eth.get_block("latest")["baseFeePerGas"] == base_fee_override


def test_reset_and_set_next_block_base_fee(fork_mainnet_full: AnvilFork):
    base_fee_override = 69 * 10**9

    starting_block = fork_mainnet_full.w3.eth.get_block_number()
    fork_mainnet_full.reset(block_number=starting_block - 10)
    fork_mainnet_full.set_next_base_fee(base_fee_override)
    fork_mainnet_full.mine()
    assert fork_mainnet_full.w3.eth.get_block_number() == starting_block - 9
    assert (
        fork_mainnet_full.w3.eth.get_block(starting_block - 9)["baseFeePerGas"] == base_fee_override
    )


def test_reset_to_new_endpoint():
    fork = AnvilFork(fork_url=ETHEREUM_FULL_NODE_HTTP_URI)
    assert fork.w3.eth.chain_id == 1

    fork.reset(fork_url=BASE_FULL_NODE_HTTP_URI)
    assert fork.w3.eth.chain_id == 8453


def test_reset_to_new_transaction_hash():
    fork = AnvilFork(fork_url=ETHEREUM_ARCHIVE_NODE_HTTP_URI)
    fork.reset(
        transaction_hash="0x12167fa2a4cd676a6e740edb09427469ecb8718d84ef4d0d5819fe8b527964d6"
    )
    assert fork.w3.eth.block_number == 20987963


def test_ipc_kwargs():
    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        ipc_provider_kwargs=dict(timeout=None),
    )
    if TYPE_CHECKING:
        assert isinstance(fork.w3.provider, IPCProvider)
    assert fork.w3.provider.timeout is None


def test_balance_overrides_in_constructor():
    fake_balance = 100 * 10**18
    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        balance_overrides=[
            (VITALIK_ADDRESS, fake_balance),
        ],
    )
    assert fork.w3.eth.get_balance(VITALIK_ADDRESS) == fake_balance


def test_nonce_overrides_in_constructor():
    fake_nonce = 69
    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        nonce_overrides=[
            (VITALIK_ADDRESS, fake_nonce),
        ],
    )
    assert fork.w3.eth.get_transaction_count(VITALIK_ADDRESS) == fake_nonce


def test_bytecode_overrides_in_constructor():
    fake_address = get_checksum_address("0x6969696969696969696969696969696969696969")
    fake_bytecode = HexBytes("0x0420")

    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI, bytecode_overrides=[(fake_address, fake_bytecode)]
    )
    assert fork.w3.eth.get_code(fake_address) == fake_bytecode


def test_coinbase_override_in_constructor():
    fake_coinbase = get_checksum_address("0x6969696969696969696969696969696969696969")

    fork = AnvilFork(
        fork_url=ETHEREUM_FULL_NODE_HTTP_URI,
        coinbase=fake_coinbase,
    )
    fork.mine()
    block = fork.w3.eth.get_block("latest")
    assert block["miner"] == fake_coinbase


def test_injecting_middleware():
    fork = AnvilFork(
        fork_url="https://polygon-bor-rpc.publicnode.com",
        middlewares=[
            (web3.middleware.ExtraDataToPOAMiddleware, 0),
        ],
    )
    set_web3(fork.w3)



from typing import TYPE_CHECKING

import eth_account
import pytest

from degenbot import AnvilFork, BuilderEndpoint
from degenbot.exceptions import DegenbotValueError

if TYPE_CHECKING:
    from eth_account.signers.local import LocalAccount

BEAVERBUILD_URL = "https://rpc.beaverbuild.org"
BUILDER0X69_URL = "https://builder0x69.io"
FLASHBOTS_URL = "https://relay.flashbots.net"
RSYNCBUILDER_URL = "https://rsync-builder.xyz"
TITANBUILDER_URL = "https://rpc.titanbuilder.xyz"

BUILDER_FIXTURES = [
    "beaverbuild",
    "builder0x69",
    "flashbots",
    "rsyncbuilder",
    "titanbuilder",
]


# Taken from https://privatekeys.pw/keys/ethereum/random
SIGNER_KEY = "52661f05c0512d64e2dc681900f45996e9946856ec352b7a2950203b150dbd28"


@pytest.fixture
def beaverbuild() -> BuilderEndpoint:
    return BuilderEndpoint(
        url=BEAVERBUILD_URL,
        endpoints=["eth_sendBundle"],
    )


@pytest.fixture
def builder0x69() -> BuilderEndpoint:
    return BuilderEndpoint(
        url=BUILDER0X69_URL,
        endpoints=["eth_sendBundle"],
        # ref: https://docs.builder0x69.io/
        authentication_header_label="X-Flashbots-Signature",
    )


@pytest.fixture
def flashbots() -> BuilderEndpoint:
    return BuilderEndpoint(
        url=FLASHBOTS_URL,
        endpoints=[
            "eth_callBundle",
            "eth_sendBundle",
            "flashbots_getUserStatsV2",
            "flashbots_getBundleStatsV2",
        ],
        authentication_header_label="X-Flashbots-Signature",
    )


@pytest.fixture
def rsyncbuilder() -> BuilderEndpoint:
    return BuilderEndpoint(
        url=RSYNCBUILDER_URL,
        endpoints=[
            "eth_cancelBundle",
            "eth_sendBundle",
            "eth_sendPrivateRawTransaction",
        ],
        authentication_header_label="X-Flashbots-Signature",
    )


@pytest.fixture
def titanbuilder() -> BuilderEndpoint:
    return BuilderEndpoint(
        url=TITANBUILDER_URL,
        endpoints=["eth_sendBundle"],
        # ref: https://docs.titanbuilder.xyz/authentication
        authentication_header_label="X-Flashbots-Signature",
    )


@pytest.mark.parametrize(
    "builder_name",
    BUILDER_FIXTURES,
)
def test_create_builders(builder_name: str, request: pytest.FixtureRequest):
    builder = request.getfixturevalue(builder_name)
    assert isinstance(builder, BuilderEndpoint)


async def test_bad_url():
    with pytest.raises(DegenbotValueError):
        BuilderEndpoint(url="ws://www.google.com", endpoints=["eth_sendBundle"])


async def test_eth_call_bundle(
    flashbots: BuilderEndpoint,
    fork_mainnet_full: AnvilFork,
):
    current_block = fork_mainnet_full.w3.eth.block_number
    current_base_fee = fork_mainnet_full.w3.eth.get_block("latest")["baseFeePerGas"]
    current_block_timestamp = fork_mainnet_full.w3.eth.get_block("latest")["timestamp"]

    signer: LocalAccount = eth_account.Account.from_key(SIGNER_KEY)
    signer_address = signer.address

    transaction_1 = {
        "chainId": 1,
        "data": b"",
        "from": signer_address,
        "to": signer_address,
        "value": 1,
        "nonce": fork_mainnet_full.w3.eth.get_transaction_count(signer_address),
        "gas": 50_000,
        "maxFeePerGas": int(1.5 * current_base_fee),
        "maxPriorityFeePerGas": 0,
    }
    transaction_2 = {
        "chainId": 1,
        "data": b"",
        "from": signer_address,
        "to": signer_address,
        "value": 1,
        "nonce": fork_mainnet_full.w3.eth.get_transaction_count(signer_address) + 1,
        "gas": 50_000,
        "maxFeePerGas": int(1.5 * current_base_fee),
        "maxPriorityFeePerGas": 0,
    }
    signed_tx_1 = fork_mainnet_full.w3.eth.account.sign_transaction(
        transaction_1, SIGNER_KEY
    ).raw_transaction
    signed_tx_2 = fork_mainnet_full.w3.eth.account.sign_transaction(
        transaction_2, SIGNER_KEY
    ).raw_transaction

    response = await flashbots.call_eth_bundle(
        bundle=[signed_tx_1, signed_tx_2],
        block_number=current_block + 1,
        state_block=current_block,
        signer_key=SIGNER_KEY,
    )
    assert isinstance(response, dict)

    # Test with "latest" state block alias
    response = await flashbots.call_eth_bundle(
        bundle=[signed_tx_1, signed_tx_2],
        block_number=current_block + 1,
        state_block="latest",
        signer_key=SIGNER_KEY,
    )
    assert isinstance(response, dict)

    # Test with timestamp
    response = await flashbots.call_eth_bundle(
        bundle=[signed_tx_1, signed_tx_2],
        block_number=current_block + 1,
        block_timestamp=current_block_timestamp + 12,
        state_block=current_block,
        signer_key=SIGNER_KEY,
    )
    assert isinstance(response, dict)


async def test_eth_send_bundle(
    flashbots: BuilderEndpoint,
    fork_mainnet_full: AnvilFork,
):
    current_block = fork_mainnet_full.w3.eth.block_number
    current_base_fee = fork_mainnet_full.w3.eth.get_block("latest")["baseFeePerGas"]

    signer: LocalAccount = eth_account.Account.from_key(SIGNER_KEY)
    signer_address = signer.address

    transaction_1 = {
        "chainId": 1,
        "data": b"",
        "from": signer_address,
        "to": signer_address,
        "value": 1,
        "nonce": fork_mainnet_full.w3.eth.get_transaction_count(signer_address),
        "gas": 50_000,
        "maxFeePerGas": int(1.5 * current_base_fee),
        "maxPriorityFeePerGas": 0,
    }
    transaction_2 = {
        "chainId": 1,
        "data": b"",
        "from": signer_address,
        "to": signer_address,
        "value": 1,
        "nonce": fork_mainnet_full.w3.eth.get_transaction_count(signer_address) + 1,
        "gas": 50_000,
        "maxFeePerGas": int(1.5 * current_base_fee),
        "maxPriorityFeePerGas": 0,
    }
    signed_tx_1 = fork_mainnet_full.w3.eth.account.sign_transaction(
        transaction_1, SIGNER_KEY
    ).raw_transaction
    signed_tx_2 = fork_mainnet_full.w3.eth.account.sign_transaction(
        transaction_2, SIGNER_KEY
    ).raw_transaction

    response = await flashbots.send_eth_bundle(
        bundle=[signed_tx_1, signed_tx_2],
        block_number=current_block + 1,
        signer_key=SIGNER_KEY,
    )
    assert isinstance(response, dict)

    await flashbots.get_bundle_stats(
        bundle_hash=response["bundleHash"],
        block_number=current_block + 1,
        signer_key=SIGNER_KEY,
    )


async def test_get_user_stats(
    flashbots: BuilderEndpoint,
    fork_mainnet_full: AnvilFork,
):
    current_block = fork_mainnet_full.w3.eth.block_number
    await flashbots.get_user_stats(
        signer_key=SIGNER_KEY,
        recent_block_number=current_block,
    )



import pytest
from hexbytes import HexBytes

from degenbot import (
    AnvilFork,
    Erc20Token,
    EtherPlaceholder,
    async_connection_manager,
    get_checksum_address,
    set_web3,
)
from degenbot.constants import ZERO_ADDRESS
from degenbot.erc20.erc20 import get_token_from_database
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.erc20 import NoPriceOracle
from degenbot.types.concrete import BoundedCache

VITALIK_ADDRESS = get_checksum_address("0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045")
WETH_ADDRESS = get_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")
WBTC_ADDRESS = get_checksum_address("0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599")
CHAINLINK_WETH_PRICE_FEED = get_checksum_address("0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419")


@pytest.fixture
def wbtc(fork_mainnet_full: AnvilFork) -> Erc20Token:
    set_web3(fork_mainnet_full.w3)
    return Erc20Token(WBTC_ADDRESS)


@pytest.fixture
def weth(fork_mainnet_full: AnvilFork) -> Erc20Token:
    set_web3(fork_mainnet_full.w3)
    return Erc20Token(WETH_ADDRESS)


def test_erc20_token_database(
    fork_mainnet_full: AnvilFork,
    weth: Erc20Token,
):
    set_web3(fork_mainnet_full.w3)
    token_from_db = get_token_from_database(
        token=weth.address,
        chain_id=fork_mainnet_full.w3.eth.chain_id,
    )
    assert token_from_db is not None


def test_bad_address(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    with pytest.raises(DegenbotValueError, match="No contract deployed at this address"):
        Erc20Token(VITALIK_ADDRESS)


def test_caches(fork_mainnet_full: AnvilFork, wbtc: Erc20Token):
    fake_balance = 69_420_000
    current_block = fork_mainnet_full.w3.eth.block_number
    balance_actual = wbtc.get_balance(VITALIK_ADDRESS)
    wbtc._cached_balance[VITALIK_ADDRESS] = BoundedCache(max_items=5)
    wbtc._cached_balance[VITALIK_ADDRESS][current_block] = fake_balance
    assert wbtc.get_balance(VITALIK_ADDRESS) == fake_balance
    wbtc._cached_balance.clear()
    assert wbtc.get_balance(VITALIK_ADDRESS) == balance_actual

    current_total_supply = wbtc.get_total_supply()
    fake_supply = 69_420_000_000
    wbtc._cached_total_supply[current_block] = fake_supply
    assert wbtc.get_total_supply() == fake_supply
    wbtc._cached_total_supply.clear()
    assert wbtc.get_total_supply() == current_total_supply

    wbtc.get_approval(VITALIK_ADDRESS, VITALIK_ADDRESS)
    wbtc.get_approval(VITALIK_ADDRESS, VITALIK_ADDRESS)


def test_erc20token_comparisons(wbtc: Erc20Token, weth: Erc20Token):
    with pytest.raises(AssertionError):
        assert weth == 69

    with pytest.raises(TypeError):
        assert weth < 69

    with pytest.raises(TypeError):
        assert weth > 69

    assert weth != wbtc

    assert weth == WETH_ADDRESS
    assert weth == WETH_ADDRESS.lower()
    assert weth == WETH_ADDRESS.upper()
    assert weth == get_checksum_address(WETH_ADDRESS)
    assert weth == HexBytes(WETH_ADDRESS)
    assert weth == bytes.fromhex(WETH_ADDRESS[2:])

    assert wbtc == WBTC_ADDRESS
    assert wbtc == WBTC_ADDRESS.lower()
    assert wbtc == WBTC_ADDRESS.upper()
    assert wbtc == get_checksum_address(WBTC_ADDRESS)
    assert wbtc == HexBytes(WBTC_ADDRESS)

    assert weth > wbtc
    assert weth > WBTC_ADDRESS
    assert weth > WBTC_ADDRESS.lower()
    assert weth > WBTC_ADDRESS.upper()
    assert weth > get_checksum_address(WBTC_ADDRESS)
    assert weth > HexBytes(WBTC_ADDRESS)
    assert weth > bytes.fromhex(WBTC_ADDRESS[2:])

    assert wbtc < weth
    assert wbtc < WETH_ADDRESS
    assert wbtc < WETH_ADDRESS.lower()
    assert wbtc < WETH_ADDRESS.upper()
    assert wbtc < get_checksum_address(WETH_ADDRESS)
    assert wbtc < HexBytes(WETH_ADDRESS)
    assert wbtc < bytes.fromhex(WETH_ADDRESS[2:])


def test_non_compliant_tokens(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    for token_address in [
        "0x0d88eD6E74bbFD96B831231638b66C05571e824F",
        "0x043942281890d4876D26BD98E2BB3F662635DFfb",
        "0x1da4858ad385cc377165A298CC2CE3fce0C5fD31",
        "0x9A2548335a639a58F4241b85B5Fc6c57185C428A",
        "0xC19B6A4Ac7C7Cc24459F08984Bbd09664af17bD1",
        "0xC36442b4a4522E871399CD717aBDD847Ab11FE88",
        "0xf5BF148Be50f6972124f223215478519A2787C8E",
        "0xfCf163B5C68bE47f702432F0f54B58Cd6E18D10B",
        "0x431ad2ff6a9C365805eBaD47Ee021148d6f7DBe0",
        "0x89d24A6b4CcB1B6fAA2625fE562bDD9a23260359",
        "0xEB9951021698B42e4399f9cBb6267Aa35F82D59D",
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
    ]:
        Erc20Token(token_address)


def test_erc20token_with_price_feed(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    weth = Erc20Token(address=WETH_ADDRESS, oracle_address=CHAINLINK_WETH_PRICE_FEED)
    _ = weth.price


def test_erc20token_without_price_feed(
    fork_mainnet_full: AnvilFork,
    weth: Erc20Token,
):
    set_web3(fork_mainnet_full.w3)

    with pytest.raises(NoPriceOracle):
        _ = weth.price


def test_erc20token_functions(fork_mainnet_full: AnvilFork, weth: Erc20Token):
    set_web3(fork_mainnet_full.w3)
    weth.get_total_supply()
    weth.get_approval(VITALIK_ADDRESS, weth.address)
    weth.get_balance(VITALIK_ADDRESS)


async def test_async_erc20_functions(fork_mainnet_full: AnvilFork, weth: Erc20Token):
    async with fork_mainnet_full.async_w3 as w3:
        await async_connection_manager.register_web3(w3)
        await weth.get_total_supply_async()
        await weth.get_approval_async(VITALIK_ADDRESS, weth.address)
        await weth.get_balance_async(VITALIK_ADDRESS)


def test_ether_placeholder(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    ether = EtherPlaceholder(ZERO_ADDRESS)

    fake_balance = 69_420_000
    current_block = fork_mainnet_full.w3.eth.block_number
    balance_actual = ether.get_balance(VITALIK_ADDRESS)
    ether._cached_balance[VITALIK_ADDRESS] = BoundedCache(max_items=5)
    ether._cached_balance[VITALIK_ADDRESS][current_block] = fake_balance
    assert ether.get_balance(VITALIK_ADDRESS) == fake_balance
    ether._cached_balance.clear()
    assert ether.get_balance(VITALIK_ADDRESS) == balance_actual



from decimal import Decimal

SCALING_FACTOR = Decimal(1 * 10**18)


def bn(x: int | Decimal) -> int:
    return int(x)


def fp(x: int | Decimal) -> int:
    return bn(to_fp(x))


def to_fp(x: int | Decimal) -> Decimal:
    return Decimal(x) * SCALING_FACTOR


def from_fp(x: int | Decimal) -> Decimal:
    return Decimal(x) / SCALING_FACTOR
